import os
import shutil
import tempfile

import click

from clarifai.cli.base import cli
from clarifai.utils.cli import validate_context
from clarifai.utils.constants import (
    DEFAULT_LOCAL_DEV_APP_ID,
    DEFAULT_LOCAL_DEV_COMPUTE_CLUSTER_CONFIG,
    DEFAULT_LOCAL_DEV_COMPUTE_CLUSTER_ID,
    DEFAULT_LOCAL_DEV_DEPLOYMENT_ID,
    DEFAULT_LOCAL_DEV_MODEL_ID,
    DEFAULT_LOCAL_DEV_MODEL_TYPE,
    DEFAULT_LOCAL_DEV_NODEPOOL_CONFIG,
    DEFAULT_LOCAL_DEV_NODEPOOL_ID,
)
from clarifai.utils.logging import logger
from clarifai.utils.misc import (
    clone_github_repo,
    format_github_repo_url,
)


@cli.group(
    ['model'], context_settings={'max_content_width': shutil.get_terminal_size().columns - 10}
)
def model():
    """Manage & Develop Models: init, download-checkpoints, signatures, upload\n
    Run & Test Models Locally: local-runner, local-grpc, local-test\n
    Model Inference: list, predict"""


@model.command()
@click.argument(
    "model_path",
    type=click.Path(),
    required=False,
    default=".",
)
@click.option(
    '--model-type-id',
    type=click.Choice(['mcp', 'openai'], case_sensitive=False),
    required=False,
    help='Model type: "mcp" for MCPModelClass, "openai" for OpenAIModelClass, or leave empty for default ModelClass.',
)
@click.option(
    '--github-pat',
    required=False,
    help='GitHub Personal Access Token for authentication when cloning private repositories.',
)
@click.option(
    '--github-repo',
    required=False,
    help='GitHub repository URL or "user/repo" format to clone a repository from. If provided, the entire repository contents will be copied to the target directory instead of using default templates.',
)
@click.option(
    '--branch',
    required=False,
    help='Git branch to clone from the GitHub repository. If not specified, the default branch will be used.',
)
def init(model_path, model_type_id, github_pat, github_repo, branch):
    """Initialize a new model directory structure.

    Creates the following structure in the specified directory:
    ├── 1/
    │   └── model.py
    ├── requirements.txt
    └── config.yaml

    If --github-repo is provided, the entire repository contents will be copied to the target
    directory instead of using default templates. The --github-pat option can be used for authentication
    when cloning private repositories. The --branch option can be used to specify a specific
    branch to clone from.

    MODEL_PATH: Path where to create the model directory structure. If not specified, the current directory is used by default.
    """
    # Resolve the absolute path
    model_path = os.path.abspath(model_path)

    # Create the model directory if it doesn't exist
    os.makedirs(model_path, exist_ok=True)

    # Handle GitHub repository cloning if provided
    if github_repo:
        logger.info(f"Initializing model from GitHub repository: {github_repo}")

        # Check if it's a local path or normalize the GitHub repo URL
        if os.path.exists(github_repo):
            repo_url = github_repo
        else:
            repo_url = format_github_repo_url(github_repo)

        # Create a temporary directory for cloning
        with tempfile.TemporaryDirectory() as temp_dir:
            clone_dir = os.path.join(temp_dir, "repo")

            # Clone the repository
            if not clone_github_repo(repo_url, clone_dir, github_pat, branch):
                logger.error(
                    "Failed to clone repository. Falling back to template-based initialization."
                )
                github_repo = None  # Fall back to template mode
            else:
                # Copy the entire repository content to target directory (excluding .git)
                for item in os.listdir(clone_dir):
                    if item == '.git':
                        continue

                    source_path = os.path.join(clone_dir, item)
                    target_path = os.path.join(model_path, item)

                    if os.path.isdir(source_path):
                        shutil.copytree(source_path, target_path, dirs_exist_ok=True)
                    else:
                        shutil.copy2(source_path, target_path)

                logger.info("Model initialization complete with GitHub repository")
                logger.info("Next steps:")
                logger.info("1. Review the model configuration")
                logger.info("2. Install any required dependencies manually")
                logger.info("3. Test the model locally using 'clarifai model local-test'")
                return

    # Fall back to template-based initialization if no GitHub repo or if GitHub repo failed
    if not github_repo:
        from clarifai.cli.templates.model_templates import (
            get_config_template,
            get_model_template,
            get_requirements_template,
        )

        # Create the 1/ subdirectory
        model_version_dir = os.path.join(model_path, "1")
        os.makedirs(model_version_dir, exist_ok=True)

        # Create model.py
        model_py_path = os.path.join(model_version_dir, "model.py")
        if os.path.exists(model_py_path):
            logger.warning(f"File {model_py_path} already exists, skipping...")
        else:
            model_template = get_model_template(model_type_id)
            with open(model_py_path, 'w') as f:
                f.write(model_template)
            logger.info(f"Created {model_py_path}")

        # Create requirements.txt
        requirements_path = os.path.join(model_path, "requirements.txt")
        if os.path.exists(requirements_path):
            logger.warning(f"File {requirements_path} already exists, skipping...")
        else:
            requirements_template = get_requirements_template(model_type_id)
            with open(requirements_path, 'w') as f:
                f.write(requirements_template)
            logger.info(f"Created {requirements_path}")

        # Create config.yaml
        config_path = os.path.join(model_path, "config.yaml")
        if os.path.exists(config_path):
            logger.warning(f"File {config_path} already exists, skipping...")
        else:
            config_model_type_id = "text-to-text"  # default

            config_template = get_config_template(config_model_type_id)
            with open(config_path, 'w') as f:
                f.write(config_template)
            logger.info(f"Created {config_path}")

        logger.info(f"Model initialization complete in {model_path}")
        logger.info("Next steps:")
        logger.info("1. Search for '# TODO: please fill in' comments in the generated files")
        logger.info("2. Update the model configuration in config.yaml")
        logger.info("3. Add your model dependencies to requirements.txt")
        logger.info("4. Implement your model logic in 1/model.py")


@model.command(help="Upload a trained model.")
@click.argument("model_path", type=click.Path(exists=True), required=False, default=".")
@click.option(
    '--stage',
    required=False,
    type=click.Choice(['runtime', 'build', 'upload'], case_sensitive=True),
    default="upload",
    show_default=True,
    help='The stage we are calling download checkpoints from. Typically this would "upload" and will download checkpoints if config.yaml checkpoints section has when set to "upload". Other options include "runtime" to be used in load_model or "upload" to be used during model upload. Set this stage to whatever you have in config.yaml to force downloading now.',
)
@click.option(
    '--skip_dockerfile',
    is_flag=True,
    help='Flag to skip generating a dockerfile so that you can manually edit an already created dockerfile.',
)
@click.pass_context
def upload(ctx, model_path, stage, skip_dockerfile):
    """Upload a model to Clarifai.

    MODEL_PATH: Path to the model directory. If not specified, the current directory is used by default.
    """
    from clarifai.runners.models.model_builder import upload_model

    validate_context(ctx)
    upload_model(
        model_path,
        stage,
        skip_dockerfile,
        pat=ctx.obj.current.pat,
        base_url=ctx.obj.current.api_base,
    )


@model.command(help="Download model checkpoint files.")
@click.argument(
    "model_path",
    type=click.Path(exists=True),
    required=False,
    default=".",
)
@click.option(
    '--out_path',
    type=click.Path(exists=False),
    required=False,
    default=None,
    help='Option path to write the checkpoints to. This will place them in {out_path}/1/checkpoints If not provided it will default to {model_path}/1/checkpoints where the config.yaml is read.',
)
@click.option(
    '--stage',
    required=False,
    type=click.Choice(['runtime', 'build', 'upload'], case_sensitive=True),
    default="build",
    show_default=True,
    help='The stage we are calling download checkpoints from. Typically this would be in the build stage which is the default. Other options include "runtime" to be used in load_model or "upload" to be used during model upload. Set this stage to whatever you have in config.yaml to force downloading now.',
)
def download_checkpoints(model_path, out_path, stage):
    """Download checkpoints from external source to local model_path

    MODEL_PATH: Path to the model directory. If not specified, the current directory is used by default.
    """

    from clarifai.runners.models.model_builder import ModelBuilder

    builder = ModelBuilder(model_path, download_validation_only=True)
    builder.download_checkpoints(stage=stage, checkpoint_path_override=out_path)


@model.command(help="Generate model method signatures.")
@click.argument(
    "model_path",
    type=click.Path(exists=True),
    required=False,
    default=".",
)
@click.option(
    '--out_path',
    type=click.Path(exists=False),
    required=False,
    default=None,
    help='Path to write the method signature defitions to. If not provided, use stdout.',
)
def signatures(model_path, out_path):
    """Generate method signatures for the model.

    MODEL_PATH: Path to the model directory. If not specified, the current directory is used by default.
    """

    from clarifai.runners.models.model_builder import ModelBuilder

    builder = ModelBuilder(model_path, download_validation_only=True)
    signatures = builder.method_signatures_yaml()
    if out_path:
        with open(out_path, 'w') as f:
            f.write(signatures)
    else:
        click.echo(signatures)


@model.command(name="local-test", help="Execute all model unit tests locally.")
@click.argument(
    "model_path",
    type=click.Path(exists=True),
    required=False,
    default=".",
)
@click.option(
    '--mode',
    type=click.Choice(['env', 'container'], case_sensitive=False),
    default='env',
    show_default=True,
    help='Specify how to test the model locally: "env" for virtual environment or "container" for Docker container. Defaults to "env".',
)
@click.option(
    '--keep_env',
    is_flag=True,
    help='Keep the virtual environment after testing the model locally (applicable for virtualenv mode). Defaults to False.',
)
@click.option(
    '--keep_image',
    is_flag=True,
    help='Keep the Docker image after testing the model locally (applicable for container mode). Defaults to False.',
)
@click.option(
    '--skip_dockerfile',
    is_flag=True,
    help='Flag to skip generating a dockerfile so that you can manually edit an already created dockerfile. Apply for `--mode conatainer`.',
)
def test_locally(model_path, keep_env=False, keep_image=False, mode='env', skip_dockerfile=False):
    """Test model locally.

    MODEL_PATH: Path to the model directory. If not specified, the current directory is used by default.
    """
    try:
        from clarifai.runners.models import model_run_locally

        if mode == 'env' and keep_image:
            raise ValueError("'keep_image' is applicable only for 'container' mode")
        if mode == 'container' and keep_env:
            raise ValueError("'keep_env' is applicable only for 'env' mode")

        if mode == "env":
            click.echo("Testing model locally in a virtual environment...")
            model_run_locally.main(model_path, run_model_server=False, keep_env=keep_env)
        elif mode == "container":
            click.echo("Testing model locally inside a container...")
            model_run_locally.main(
                model_path,
                inside_container=True,
                run_model_server=False,
                keep_image=keep_image,
                skip_dockerfile=skip_dockerfile,
            )
        click.echo("Model tested successfully.")
    except Exception as e:
        click.echo(f"Failed to test model locally: {e}", err=True)


@model.command(name="local-grpc", help="Run the model locally via a gRPC server.")
@click.argument(
    "model_path",
    type=click.Path(exists=True),
    required=False,
    default=".",
)
@click.option(
    '--port',
    '-p',
    type=int,
    default=8000,
    show_default=True,
    help="The port to host the gRPC server for running the model locally. Defaults to 8000.",
)
@click.option(
    '--mode',
    type=click.Choice(['env', 'container'], case_sensitive=False),
    default='env',
    show_default=True,
    help='Specifies how to run the model: "env" for virtual environment or "container" for Docker container. Defaults to "env".',
)
@click.option(
    '--keep_env',
    is_flag=True,
    help='Keep the virtual environment after testing the model locally (applicable for virtualenv mode). Defaults to False.',
)
@click.option(
    '--keep_image',
    is_flag=True,
    help='Keep the Docker image after testing the model locally (applicable for container mode). Defaults to False.',
)
@click.option(
    '--skip_dockerfile',
    is_flag=True,
    help='Flag to skip generating a dockerfile so that you can manually edit an already created dockerfile. Apply for `--mode conatainer`.',
)
def run_locally(model_path, port, mode, keep_env, keep_image, skip_dockerfile=False):
    """Run the model locally and start a gRPC server to serve the model.

    MODEL_PATH: Path to the model directory. If not specified, the current directory is used by default.
    """
    try:
        from clarifai.runners.models import model_run_locally

        if mode == 'env' and keep_image:
            raise ValueError("'keep_image' is applicable only for 'container' mode")
        if mode == 'container' and keep_env:
            raise ValueError("'keep_env' is applicable only for 'env' mode")

        if mode == "env":
            click.echo("Running model locally in a virtual environment...")
            model_run_locally.main(model_path, run_model_server=True, keep_env=keep_env, port=port)
        elif mode == "container":
            click.echo("Running model locally inside a container...")
            model_run_locally.main(
                model_path,
                inside_container=True,
                run_model_server=True,
                port=port,
                keep_image=keep_image,
                skip_dockerfile=skip_dockerfile,
            )
        click.echo(f"Model server started locally from {model_path} in {mode} mode.")
    except Exception as e:
        click.echo(f"Failed to starts model server locally: {e}", err=True)


@model.command(name="local-runner", help="Run the model locally for dev, debug, or local compute.")
@click.argument(
    "model_path",
    type=click.Path(exists=True),
    required=False,
    default=".",
)
@click.pass_context
def local_dev(ctx, model_path):
    """Run the model as a local dev runner to help debug your model connected to the API or to
    leverage local compute resources manually. This relies on many variables being present in the env
    of the currently selected context. If they are not present then default values will be used to
    ease the setup of a local dev runner and your context yaml will be updated in place. The required
    env vars are:

    \b
      CLARIFAI_PAT:

    \b
      # for where the model that represents the local runner should be:
    \b
      CLARIFAI_USER_ID:
      CLARIFAI_APP_ID:
      CLARIFAI_MODEL_ID:

    \b
      # for where the local dev runner should be in a compute cluser
      # note the user_id of the compute cluster is the same as the user_id of the model.

    \b
      CLARIFAI_COMPUTE_CLUSTER_ID:
      CLARIFAI_NODEPOOL_ID:

      # The following will be created in your context since it's generated by the API

      CLARIFAI_RUNNER_ID:


    Additionally using the provided model path, if the config.yaml file does not contain the model
    information that matches the above CLARIFAI_USER_ID, CLARIFAI_APP_ID, CLARIFAI_MODEL_ID then the
    config.yaml will be updated to include the model information. This is to ensure that the model
    that starts up in the local dev runner is the same as the one you intend to call in the API.

    MODEL_PATH: Path to the model directory. If not specified, the current directory is used by default.
    """
    from clarifai.client.user import User
    from clarifai.runners.models.model_builder import ModelBuilder
    from clarifai.runners.server import serve

    validate_context(ctx)
    logger.info("Checking setup for local development runner...")
    logger.info(f"Current context: {ctx.obj.current.name}")
    user_id = ctx.obj.current.user_id
    user = User(user_id=user_id, pat=ctx.obj.current.pat, base_url=ctx.obj.current.api_base)
    logger.info(f"Current user_id: {user_id}")
    logger.debug("Checking if a local dev compute cluster exists...")

    # see if ctx has CLARIFAI_COMPUTE_CLUSTER_ID, if not use default
    try:
        compute_cluster_id = ctx.obj.current.compute_cluster_id
    except AttributeError:
        compute_cluster_id = DEFAULT_LOCAL_DEV_COMPUTE_CLUSTER_ID
    logger.info(f"Current compute_cluster_id: {compute_cluster_id}")

    try:
        compute_cluster = user.compute_cluster(compute_cluster_id)
        if compute_cluster.cluster_type != 'local-dev':
            raise ValueError(
                f"Compute cluster {user_id}/{compute_cluster_id} is not a local-dev compute cluster. Please create a local-dev compute cluster."
            )
        try:
            compute_cluster_id = ctx.obj.current.compute_cluster_id
        except AttributeError:  # doesn't exist in context but does in API then update the context.
            ctx.obj.current.CLARIFAI_COMPUTE_CLUSTER_ID = compute_cluster.id
            ctx.obj.to_yaml()  # save to yaml file.
    except ValueError:
        raise
    except Exception as e:
        logger.info(f"Failed to get compute cluster with ID {compute_cluster_id}: {e}")
        y = input(
            f"Compute cluster not found. Do you want to create a new compute cluster {user_id}/{compute_cluster_id}? (y/n): "
        )
        if y.lower() != 'y':
            raise click.Abort()
        # Create a compute cluster with default configuration for local dev.
        compute_cluster = user.create_compute_cluster(
            compute_cluster_id=compute_cluster_id,
            compute_cluster_config=DEFAULT_LOCAL_DEV_COMPUTE_CLUSTER_CONFIG,
        )
        ctx.obj.current.CLARIFAI_COMPUTE_CLUSTER_ID = compute_cluster_id
        ctx.obj.to_yaml()  # save to yaml file.

    # Now check if there is a nodepool created in this compute cluser
    try:
        nodepool_id = ctx.obj.current.nodepool_id
    except AttributeError:
        nodepool_id = DEFAULT_LOCAL_DEV_NODEPOOL_ID
    logger.info(f"Current nodepool_id: {nodepool_id}")

    try:
        nodepool = compute_cluster.nodepool(nodepool_id)
        try:
            nodepool_id = ctx.obj.current.nodepool_id
        except AttributeError:  # doesn't exist in context but does in API then update the context.
            ctx.obj.current.CLARIFAI_NODEPOOL_ID = nodepool.id
            ctx.obj.to_yaml()  # save to yaml file.
    except Exception as e:
        logger.info(f"Failed to get nodepool with ID {nodepool_id}: {e}")
        y = input(
            f"Nodepool not found. Do you want to create a new nodepool {user_id}/{compute_cluster_id}/{nodepool_id}? (y/n): "
        )
        if y.lower() != 'y':
            raise click.Abort()
        nodepool = compute_cluster.create_nodepool(
            nodepool_config=DEFAULT_LOCAL_DEV_NODEPOOL_CONFIG, nodepool_id=nodepool_id
        )
        ctx.obj.current.CLARIFAI_NODEPOOL_ID = nodepool_id
        ctx.obj.to_yaml()  # save to yaml file.

    logger.debug("Checking if model is created to call for local development...")
    # see if ctx has CLARIFAI_APP_ID, if not use default
    try:
        app_id = ctx.obj.current.app_id
    except AttributeError:
        app_id = DEFAULT_LOCAL_DEV_APP_ID
    logger.info(f"Current app_id: {app_id}")

    try:
        app = user.app(app_id)
        try:
            app_id = ctx.obj.current.app_id
        except AttributeError:  # doesn't exist in context but does in API then update the context.
            ctx.obj.current.CLARIFAI_APP_ID = app.id
            ctx.obj.to_yaml()  # save to yaml file.
    except Exception as e:
        logger.info(f"Failed to get app with ID {app_id}: {e}")
        y = input(f"App not found. Do you want to create a new app {user_id}/{app_id}? (y/n): ")
        if y.lower() != 'y':
            raise click.Abort()
        app = user.create_app(app_id)
        ctx.obj.current.CLARIFAI_APP_ID = app_id
        ctx.obj.to_yaml()  # save to yaml file.

    # Within this app we now need a model to call as the local dev runner.
    try:
        model_id = ctx.obj.current.model_id
    except AttributeError:
        model_id = DEFAULT_LOCAL_DEV_MODEL_ID
    logger.info(f"Current model_id: {model_id}")

    try:
        model = app.model(model_id)
        try:
            model_id = ctx.obj.current.model_id
        except AttributeError:  # doesn't exist in context but does in API then update the context.
            ctx.obj.current.CLARIFAI_MODEL_ID = model.id
            ctx.obj.to_yaml()  # save to yaml file.
    except Exception as e:
        logger.info(f"Failed to get model with ID {model_id}: {e}")
        y = input(
            f"Model not found. Do you want to create a new model {user_id}/{app_id}/models/{model_id}? (y/n): "
        )
        if y.lower() != 'y':
            raise click.Abort()
        try:
            model_type_id = ctx.obj.current.model_type_id
        except AttributeError:
            model_type_id = DEFAULT_LOCAL_DEV_MODEL_TYPE

        model = app.create_model(model_id, model_type_id=model_type_id)
        ctx.obj.current.CLARIFAI_MODEL_TYPE_ID = model_type_id
        ctx.obj.current.CLARIFAI_MODEL_ID = model_id
        ctx.obj.to_yaml()  # save to yaml file.

    # Now we need to create a version for the model if no version exists. Only need one version that
    # mentions it's a local dev runner.
    model_versions = [v for v in model.list_versions()]
    if len(model_versions) == 0:
        logger.info("No model versions found. Creating a new version for local dev runner.")
        version = model.create_version(pretrained_model_config={"local_dev": True}).model_version
        logger.info(f"Created model version {version.id}")
    else:
        version = model_versions[0].model_version

    logger.info(f"Current model version {version.id}")

    worker = {
        "model": {
            "id": f"{model.id}",
            "model_version": {
                "id": f"{version.id}",
            },
            "user_id": f"{user_id}",
            "app_id": f"{app_id}",
        },
    }

    try:
        # if it's already in our context then we'll re-use the same one.
        # note these are UUIDs, we cannot provide a runner ID.
        runner_id = ctx.obj.current.runner_id

        try:
            runner = nodepool.runner(runner_id)
            # ensure the deployment is using the latest version.
            if runner.worker.model.model_version.id != version.id:
                nodepool.delete_runners([runner_id])
                raise AttributeError("Deleted runner that was for an old model version ID.")
        except Exception as e:
            raise AttributeError("Runner not found in nodepool.") from e
    except AttributeError:
        logger.info(
            f"Create the local dev runner tying this\n  {user_id}/{app_id}/models/{model.id} model (version: {version.id}) to the\n  {user_id}/{compute_cluster_id}/{nodepool_id} nodepool."
        )
        runner = nodepool.create_runner(
            runner_config={
                "runner": {
                    "description": "Local dev runner for model testing",
                    "worker": worker,
                    "num_replicas": 1,
                }
            }
        )
        runner_id = runner.id
        ctx.obj.current.CLARIFAI_RUNNER_ID = runner.id
        ctx.obj.to_yaml()

    logger.info(f"Current runner_id: {runner_id}")

    # To make it easier to call the model without specifying a runner selector
    # we will also create a deployment tying the model to the nodepool.
    try:
        deployment_id = ctx.obj.current.deployment_id
    except AttributeError:
        deployment_id = DEFAULT_LOCAL_DEV_DEPLOYMENT_ID
    try:
        deployment = nodepool.deployment(deployment_id)
        # ensure the deployment is using the latest version.
        if deployment.worker.model.model_version.id != version.id:
            nodepool.delete_deployments([deployment_id])
            raise Exception("Deleted deployment that was for an old model version ID.")
        try:
            deployment_id = ctx.obj.current.deployment_id
        except AttributeError:  # doesn't exist in context but does in API then update the context.
            ctx.obj.current.CLARIFAI_DEPLOYMENT_ID = deployment.id
            ctx.obj.to_yaml()  # save to yaml file.
    except Exception as e:
        logger.info(f"Failed to get deployment with ID {deployment_id}: {e}")
        y = input(
            f"Deployment not found. Do you want to create a new deployment {user_id}/{compute_cluster_id}/{nodepool_id}/{deployment_id}? (y/n): "
        )
        if y.lower() != 'y':
            raise click.Abort()
        nodepool.create_deployment(
            deployment_id=deployment_id,
            deployment_config={
                "deployment": {
                    "scheduling_choice": 3,  # 3 means by price
                    "worker": worker,
                    "nodepools": [
                        {
                            "id": f"{nodepool_id}",
                            "compute_cluster": {
                                "id": f"{compute_cluster_id}",
                                "user_id": f"{user_id}",
                            },
                        }
                    ],
                    "deploy_latest_version": True,
                }
            },
        )
        ctx.obj.current.CLARIFAI_DEPLOYMENT_ID = deployment_id
        ctx.obj.to_yaml()  # save to yaml file.

    logger.info(f"Current deployment_id: {deployment_id}")

    logger.info(
        f"Full url for the model: {ctx.obj.current.ui}/users/{user_id}/apps/{app_id}/models/{model.id}/versions/{version.id}"
    )

    # Now that we have all the context in ctx.obj, we need to update the config.yaml in
    # the model_path directory with the model object containing user_id, app_id, model_id, version_id
    config_file = os.path.join(model_path, 'config.yaml')
    if not os.path.exists(config_file):
        raise ValueError(
            f"config.yaml not found in {model_path}. Please ensure you are passing the correct directory."
        )
    config = ModelBuilder._load_config(config_file)
    model_type_id = config.get('model', {}).get('model_type_id', DEFAULT_LOCAL_DEV_MODEL_TYPE)
    # The config.yaml doens't match what we created above.
    if 'model' in config and model_id != config['model'].get('id'):
        logger.info(f"Current model section of config.yaml: {config.get('model', {})}")
        y = input(
            "Do you want to backup config.yaml to config.yaml.bk then update the config.yaml with the new model information? (y/n): "
        )
        if y.lower() != 'y':
            raise click.Abort()
        config = ModelBuilder._set_local_dev_model(
            config, user_id, app_id, model_id, model_type_id
        )
        ModelBuilder._backup_config(config_file)
        ModelBuilder._save_config(config_file, config)

    builder = ModelBuilder(model_path, download_validation_only=True)
    # don't mock for local dev since you need the dependencies to run the code anyways.
    method_signatures = builder.get_method_signatures(mocking=False)

    from clarifai.runners.utils import code_script

    snippet = code_script.generate_client_script(
        method_signatures,
        user_id=user_id,
        app_id=app_id,
        model_id=model_id,
        deployment_id=deployment_id,
        base_url=ctx.obj.current.api_base,
    )

    logger.info("""\n
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# About to start up the local dev runner in this terminal...
# Here is a code snippet to call this model once it start from another terminal:
""")
    logger.info(snippet)

    logger.info("Now starting the local dev runner...")

    # This reads the config.yaml from the model_path so we alter it above first.
    serve(
        model_path,
        user_id=user_id,
        compute_cluster_id=compute_cluster_id,
        nodepool_id=nodepool_id,
        runner_id=runner_id,
        base_url=ctx.obj.current.api_base,
        pat=ctx.obj.current.pat,
    )


@model.command(help="Perform a prediction using the model.")
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=False,
    help='Path to the model predict config file.',
)
@click.option('--model_id', required=False, help='Model ID of the model used to predict.')
@click.option('--user_id', required=False, help='User ID of the model used to predict.')
@click.option('--app_id', required=False, help='App ID of the model used to predict.')
@click.option('--model_url', required=False, help='Model URL of the model used to predict.')
@click.option('--file_path', required=False, help='File path of file for the model to predict')
@click.option('--url', required=False, help='URL to the file for the model to predict')
@click.option('--bytes', required=False, help='Bytes to the file for the model to predict')
@click.option('--input_type', required=False, help='Type of input')
@click.option(
    '-cc_id',
    '--compute_cluster_id',
    required=False,
    help='Compute Cluster ID to use for the model',
)
@click.option('-np_id', '--nodepool_id', required=False, help='Nodepool ID to use for the model')
@click.option(
    '-dpl_id', '--deployment_id', required=False, help='Deployment ID to use for the model'
)
@click.option(
    '--inference_params', required=False, default='{}', help='Inference parameters to override'
)
@click.option('--output_config', required=False, default='{}', help='Output config to override')
@click.pass_context
def predict(
    ctx,
    config,
    model_id,
    user_id,
    app_id,
    model_url,
    file_path,
    url,
    bytes,
    input_type,
    compute_cluster_id,
    nodepool_id,
    deployment_id,
    inference_params,
    output_config,
):
    """Predict using the given model"""
    import json

    from clarifai.client.model import Model
    from clarifai.utils.cli import from_yaml, validate_context

    validate_context(ctx)
    if config:
        config = from_yaml(config)
        (
            model_id,
            user_id,
            app_id,
            model_url,
            file_path,
            url,
            bytes,
            input_type,
            compute_cluster_id,
            nodepool_id,
            deployment_id,
            inference_params,
            output_config,
        ) = (
            config.get(k, v)
            for k, v in [
                ('model_id', model_id),
                ('user_id', user_id),
                ('app_id', app_id),
                ('model_url', model_url),
                ('file_path', file_path),
                ('url', url),
                ('bytes', bytes),
                ('input_type', input_type),
                ('compute_cluster_id', compute_cluster_id),
                ('nodepool_id', nodepool_id),
                ('deployment_id', deployment_id),
                ('inference_params', inference_params),
                ('output_config', output_config),
            ]
        )
    if (
        sum(
            [
                opt[1]
                for opt in [(model_id, 1), (user_id, 1), (app_id, 1), (model_url, 3)]
                if opt[0]
            ]
        )
        != 3
    ):
        raise ValueError(
            "Either --model_id & --user_id & --app_id or --model_url must be provided."
        )
    if compute_cluster_id or nodepool_id or deployment_id:
        if (
            sum(
                [
                    opt[1]
                    for opt in [(compute_cluster_id, 0.5), (nodepool_id, 0.5), (deployment_id, 1)]
                    if opt[0]
                ]
            )
            != 1
        ):
            raise ValueError(
                "Either --compute_cluster_id & --nodepool_id or --deployment_id must be provided."
            )
    if model_url:
        model = Model(
            url=model_url,
            pat=ctx.obj['pat'],
            base_url=ctx.obj['base_url'],
            compute_cluster_id=compute_cluster_id,
            nodepool_id=nodepool_id,
            deployment_id=deployment_id,
        )
    else:
        model = Model(
            model_id=model_id,
            user_id=user_id,
            app_id=app_id,
            pat=ctx.obj['pat'],
            base_url=ctx.obj['base_url'],
            compute_cluster_id=compute_cluster_id,
            nodepool_id=nodepool_id,
            deployment_id=deployment_id,
        )

    if inference_params:
        inference_params = json.loads(inference_params)
    if output_config:
        output_config = json.loads(output_config)

    if file_path:
        model_prediction = model.predict_by_filepath(
            filepath=file_path,
            input_type=input_type,
            inference_params=inference_params,
            output_config=output_config,
        )
    elif url:
        model_prediction = model.predict_by_url(
            url=url,
            input_type=input_type,
            inference_params=inference_params,
            output_config=output_config,
        )
    elif bytes:
        bytes = str.encode(bytes)
        model_prediction = model.predict_by_bytes(
            input_bytes=bytes,
            input_type=input_type,
            inference_params=inference_params,
            output_config=output_config,
        )  ## TO DO: Add support for input_id
    click.echo(model_prediction)


@model.command(name="list")
@click.argument(
    "user_id",
    required=False,
    default=None,
)
@click.option(
    '--app_id',
    '-a',
    type=str,
    default=None,
    show_default=True,
    help="Get all models of an app",
)
@click.pass_context
def list_model(ctx, user_id, app_id):
    """List models of user/community.

    USER_ID: User id. If not specified, the current user is used by default. Set "all" to get all public models in Clarifai platform.
    """
    from clarifai.client import User

    try:
        pat = ctx.obj.contexts["default"]["env"]["CLARIFAI_PAT"]
    except Exception as e:
        pat = None

    User(pat=pat).list_models(
        user_id=user_id, app_id=app_id, show=True, return_clarifai_model=False
    )



import io
import json
import re
import uuid
from typing import Iterable, List, Tuple, Union, get_args, get_origin

import numpy as np
from clarifai_grpc.grpc.api.resources_pb2 import Audio as AudioProto
from clarifai_grpc.grpc.api.resources_pb2 import Concept as ConceptProto
from clarifai_grpc.grpc.api.resources_pb2 import Frame as FrameProto
from clarifai_grpc.grpc.api.resources_pb2 import Image as ImageProto
from clarifai_grpc.grpc.api.resources_pb2 import Mask as MaskProto
from clarifai_grpc.grpc.api.resources_pb2 import Point as PointProto
from clarifai_grpc.grpc.api.resources_pb2 import Region as RegionProto
from clarifai_grpc.grpc.api.resources_pb2 import Text as TextProto
from clarifai_grpc.grpc.api.resources_pb2 import Video as VideoProto
from PIL import Image as PILImage

__all__ = [
    "Audio",
    "Concept",
    "Frame",
    "Image",
    "Region",
    "Text",
    "Video",
    "MessageData",
    "NamedFieldsMeta",
    "NamedFields",
    "JSON",
]


class MessageData:
    def to_proto(self):
        raise NotImplementedError

    @classmethod
    def from_proto(cls, proto):
        raise NotImplementedError

    @classmethod
    def from_value(cls, value):
        if isinstance(value, cls):
            return value
        return cls(value)

    def cast(self, python_type):
        if python_type is self.__class__:
            return self
        raise TypeError(f'Incompatible type for {self.__class__.__name__}: {python_type}')


class NamedFieldsMeta(type):
    """Metaclass to create NamedFields subclasses with __annotations__ when fields are specified."""

    def __call__(cls, *args, **kwargs):
        # Check if keyword arguments are types (used in type annotations)
        if kwargs and all(isinstance(v, type) for v in kwargs.values()):
            # Dynamically create a subclass with __annotations__
            name = f"NamedFields({', '.join(f'{k}:{v.__name__}' for k, v in kwargs.items())})"
            return type(name, (cls,), {'__annotations__': kwargs})
        else:
            # Create a normal instance for runtime data
            return super().__call__(*args, **kwargs)


class NamedFields(metaclass=NamedFieldsMeta):
    """A class that can be used to store named fields with values."""

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def __contains__(self, key):
        return key in self.__dict__

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(f'{key}={value!r}' for key, value in self.__dict__.items())})"

    def __origin__(self):
        return self

    def __args__(self):
        return list(self.keys())


class JSON:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other

    def __bool__(self):
        return bool(self.value)

    def __repr__(self):
        return f"JSON({self.value})"

    def to_json(self):
        return json.dumps(self.value)

    @classmethod
    def from_json(cls, json_str):
        return cls(json.loads(json_str))

    @classmethod
    def from_value(cls, value):
        return cls(value)

    def cast(self, python_type):
        if not isinstance(self.value, python_type):
            raise TypeError(f'Incompatible type {type(self.value)} for {python_type}')
        return self.value


class Text(MessageData):
    def __init__(self, text: str, url: str = None):
        self.text = text
        self.url = url

    def __eq__(self, other):
        if isinstance(other, Text):
            return self.text == other.text and self.url == other.url
        if isinstance(other, str):
            return self.text == other
        return False

    def __bool__(self):
        return bool(self.text) or bool(self.url)

    def __repr__(self) -> str:
        return f"Text(text={self.text!r}, url={self.url!r})"

    def to_proto(self) -> TextProto:
        return TextProto(raw=self.text or '', url=self.url or '')

    @classmethod
    def from_proto(cls, proto: TextProto) -> "Text":
        return cls(proto.raw, proto.url or None)

    @classmethod
    def from_value(cls, value):
        if isinstance(value, str):
            return cls(value)
        if isinstance(value, Text):
            return value
        if isinstance(value, dict):
            return cls(value.get('text'), value.get('url'))
        raise TypeError(f'Incompatible type for Text: {type(value)}')

    def cast(self, python_type):
        if python_type is str:
            return self.text
        if python_type is Text:
            return self
        raise TypeError(f'Incompatible type for Text: {python_type}')


class Concept(MessageData):
    def __init__(self, name: str, value: float = 1.0, id: str = None):
        self.name = name
        self.value = value
        self.id = id or Concept._concept_name_to_id(name)

    @staticmethod
    def _concept_name_to_id(name: str):
        _name = re.sub(r'[^a-zA-Z0-9\s]', '', name)
        _name = _name.replace(' ', '-')
        return _name

    def __repr__(self) -> str:
        return f"Concept(id={self.id!r}, name={self.name!r}, value={self.value})"

    def to_proto(self):
        return ConceptProto(id=self.id, name=self.name, value=self.value)

    @classmethod
    def from_proto(cls, proto: ConceptProto) -> "Concept":
        return cls(id=proto.id, name=proto.name, value=proto.value)


class Region(MessageData):
    def __init__(
        self,
        proto_region: RegionProto = None,
        box: List[float] = None,
        concepts: List[Concept] = None,
        mask: Union['Image', PILImage.Image, np.ndarray] = None,
        point: Tuple[float, float, float] = None,
        track_id: str = None,
        text: str = None,
        id: str = None,
    ):
        if proto_region is None:
            self.proto = RegionProto()
            # use setters for init vals
            if box:
                self.box = box
            if concepts:
                self.concepts = concepts
            if mask:
                self.mask = mask
            if point:
                self.point = point
            if track_id:
                self.track_id = track_id
            if text:
                self.text = text
            self.id = id if id is not None else uuid.uuid4().hex
        elif isinstance(proto_region, RegionProto):
            self.proto = proto_region
        else:
            raise TypeError(
                f"Expected type {RegionProto.__name__}, but got type {type(proto_region).__name__}"
            )

    @property
    def id(self):
        return self.proto.id

    @id.setter
    def id(self, value: str):
        self.proto.id = value

    @property
    def text(self):
        return self.proto.data.text.raw

    @text.setter
    def text(self, value: str):
        self.proto.data.text.raw = value

    @property
    def box(self) -> List[float]:
        """[xmin, ymin, xmax, ymax]"""
        bbox = self.proto.region_info.bounding_box
        # x1, y1, x2, y2
        return [bbox.left_col, bbox.top_row, bbox.right_col, bbox.bottom_row]

    @box.setter
    def box(self, value: List[float]):
        if (
            isinstance(value, list)
            and len(value) == 4
            and all(isinstance(val, (int, float)) for val in value)
        ):
            bbox = self.proto.region_info.bounding_box
            bbox.left_col, bbox.top_row, bbox.right_col, bbox.bottom_row = value
        else:
            raise TypeError(f"Expected a list of 4 float values for 'box', but got: {value}")

    @property
    def concepts(self) -> List[Concept]:
        return [Concept.from_proto(proto) for proto in self.proto.data.concepts]

    @concepts.setter
    def concepts(self, value: List[Concept]):
        if isinstance(value, list) and all(isinstance(concept, Concept) for concept in value):
            self.proto.data.concepts.extend([concept.to_proto() for concept in value])
        else:
            raise TypeError(f"Expected a list of 'Concept' for 'concepts', but got: {value}")

    @property
    def mask(self) -> 'Image':
        return Image.from_proto(self.proto.region_info.mask.image)

    @mask.setter
    def mask(self, value: Union['Image', PILImage.Image, np.ndarray]):
        if isinstance(value, PILImage.Image):
            value = Image.from_pil(value)
        elif isinstance(value, np.ndarray):
            value = Image.from_numpy(value)
        elif not isinstance(value, Image):
            raise TypeError(
                f"Expected one of types ['Image', PIL.Image.Image, numpy.ndarray] got '{type(value).__name__}'"
            )
        self.proto.region_info.mask.CopyFrom(MaskProto(image=value.to_proto()))

    @property
    def track_id(self) -> str:
        return self.proto.track_id

    @track_id.setter
    def track_id(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"Expected str for track_id, got {type(value).__name__}")
        self.proto.track_id = value

    @property
    def point(self) -> Tuple[float, float, float]:
        """[x,y,z]"""
        point = self.proto.region_info.point
        return point.col, point.row, point.z

    @point.setter
    def point(self, value: Tuple[float, float, float]):
        if not isinstance(value, Iterable):
            raise TypeError(f"Expected a tuple of floats, got {type(value).__name__}")
        value = tuple(value)
        if len(value) != 3:
            raise ValueError(f"Expected 3 elements, got {len(value)}")
        if not all(isinstance(v, float) for v in value):
            raise TypeError("All elements must be floats")
        x, y, z = value
        point_proto = PointProto(col=x, row=y, z=z)
        self.proto.region_info.point.CopyFrom(point_proto)

    def __repr__(self) -> str:
        return f"Region(id={self.id},box={self.box or []}, concepts={self.concepts or None}, point={self.point or None}, mask={self.mask or None}, track_id={self.track_id or None})"

    def to_proto(self) -> RegionProto:
        return self.proto

    @classmethod
    def from_proto(cls, proto: RegionProto) -> "Region":
        return cls(proto)


class Image(MessageData):
    def __init__(self, proto_image: ImageProto = None, url: str = None, bytes: bytes = None):
        if proto_image is None:
            proto_image = ImageProto()
        self.proto = proto_image
        # use setters for init vals
        if url:
            self.url = url
        if bytes:
            self.bytes = bytes

    @property
    def url(self) -> str:
        return self.proto.url

    @url.setter
    def url(self, value: str):
        self.proto.url = value

    @property
    def bytes(self) -> bytes:
        return self.proto.base64

    @bytes.setter
    def bytes(self, value: bytes):
        self.proto.base64 = value

    def __repr__(self) -> str:
        attrs = []
        if self.url:
            attrs.append(f"url={self.url!r}")
        if self.bytes:
            attrs.append(f"bytes=<{len(self.bytes)} bytes>")
        return f"Image({', '.join(attrs)})"

    @classmethod
    def from_url(cls, url: str) -> "Image":
        proto_image = ImageProto(url=url)
        return cls(proto_image)

    @classmethod
    def from_pil(cls, pil_image: PILImage.Image, img_format="PNG") -> "Image":
        with io.BytesIO() as output:
            pil_image.save(output, format=img_format)
            image_bytes = output.getvalue()
        proto_image = ImageProto(base64=image_bytes)
        return cls(proto_image)

    @classmethod
    def from_numpy(cls, numpy_image: np.ndarray, img_format="PNG") -> "Image":
        pil_image = PILImage.fromarray(numpy_image)
        with io.BytesIO() as output:
            pil_image.save(output, format=img_format)
            image_bytes = output.getvalue()
        proto_image = ImageProto(base64=image_bytes)
        return cls(proto_image)

    def to_pil(self) -> PILImage.Image:
        if not self.proto.base64:
            raise ValueError("Image has no bytes")
        return PILImage.open(io.BytesIO(self.proto.base64))

    def to_base64_str(self) -> str:
        if not self.proto.base64:
            raise ValueError("Image has no bytes")
        if isinstance(self.proto.base64, str):
            return self.proto.base64
        if isinstance(self.proto.base64, bytes):
            try:
                # trying direct decode (if already a base64 bytes)
                return self.proto.base64.decode('utf-8')
            except UnicodeDecodeError:
                import base64

                return base64.b64encode(self.proto.base64).decode('utf-8')
        else:
            raise TypeError("Expected str or bytes for Image.base64")

    def to_numpy(self) -> np.ndarray:
        return np.asarray(self.to_pil())

    def to_proto(self) -> ImageProto:
        return self.proto

    @classmethod
    def from_proto(cls, proto: ImageProto) -> "Image":
        return cls(proto)

    @classmethod
    def from_value(cls, value):
        if isinstance(value, PILImage.Image):
            return cls.from_pil(value)
        if isinstance(value, Image):
            return value
        raise TypeError(f'Incompatible type for Image: {type(value)}')

    def cast(self, python_type):
        if python_type is Image:
            return self
        if python_type is PILImage.Image or python_type is PILImage:
            return self.to_pil()
        if python_type is np.ndarray or get_origin(python_type) is np.ndarray:
            return self.to_numpy()
        raise TypeError(f'Incompatible type for Image: {python_type}')


class Audio(MessageData):
    def __init__(self, proto_audio: AudioProto = None, url: str = None, bytes: bytes = None):
        if proto_audio is None:
            proto_audio = AudioProto()
        self.proto = proto_audio

        if url:
            self.url = url
        if bytes:
            self.bytes = bytes

    @property
    def url(self) -> str:
        return self.proto.url

    @url.setter
    def url(self, value: str):
        self.proto.url = value

    @property
    def bytes(self) -> bytes:
        return self.proto.base64

    @bytes.setter
    def bytes(self, value: bytes):
        self.proto.base64 = value

    @classmethod
    def from_url(cls, url: str) -> "Audio":
        proto_audio = AudioProto(url=url)
        return cls(proto_audio)

    def __repr__(self) -> str:
        attrs = []
        if self.url:
            attrs.append(f"url={self.url!r}")
        if self.bytes:
            attrs.append(f"bytes=<{len(self.bytes)} bytes>")
        return f"Audio({', '.join(attrs)})"

    def to_proto(self) -> AudioProto:
        return self.proto

    def to_base64_str(self) -> str:
        if not self.proto.base64:
            raise ValueError("Audio has no bytes")
        if isinstance(self.proto.base64, str):
            return self.proto.base64
        if isinstance(self.proto.base64, bytes):
            try:
                # trying direct decode (if already a base64 bytes)
                return self.proto.base64.decode('utf-8')
            except UnicodeDecodeError:
                import base64

                return base64.b64encode(self.proto.base64).decode('utf-8')
        else:
            raise TypeError("Expected str or bytes for Audio.base64")

    @classmethod
    def from_proto(cls, proto: AudioProto) -> "Audio":
        return cls(proto)


class Frame(MessageData):
    def __init__(
        self,
        proto_frame: FrameProto = None,
        image: Image = None,
        regions: List[Region] = None,
        time: float = None,
        index: int = None,
    ):
        if proto_frame is None:
            proto_frame = FrameProto()
        self.proto = proto_frame
        # use setters for init vals
        if image:
            self.image = image
        if regions:
            self.regions = regions
        if time:
            self.time = time
        if index:
            self.index = index

    @property
    def time(self) -> float:
        # TODO: time is a uint32, so this will overflow at 49.7 days
        # we should be using double or uint64 in the proto instead
        return self.proto.frame_info.time / 1000.0

    @time.setter
    def time(self, value: float):
        self.proto.frame_info.time = int(value * 1000)

    @property
    def index(self) -> int:
        return self.proto.frame_info.index

    @index.setter
    def index(self, value: int):
        self.proto.frame_info.index = value

    @property
    def image(self) -> Image:
        return Image.from_proto(self.proto.data.image)

    @image.setter
    def image(self, value: Image):
        self.proto.data.image.CopyFrom(value.to_proto())

    @property
    def regions(self) -> List[Region]:
        return [Region(region) for region in self.proto.data.regions]

    @regions.setter
    def regions(self, value: List[Region]):
        self.proto.data.regions.extend([region.proto for region in value])

    def to_proto(self) -> FrameProto:
        return self.proto

    @classmethod
    def from_proto(cls, proto: FrameProto) -> "Frame":
        return cls(proto)


class Video(MessageData):
    def __init__(self, proto_video: VideoProto = None, url: str = None, bytes: bytes = None):
        if proto_video is None:
            proto_video = VideoProto()
        self.proto = proto_video

        if url:
            self.url = url
        if bytes:
            self.bytes = bytes

    @property
    def url(self) -> str:
        return self.proto.url

    @url.setter
    def url(self, value: str):
        self.proto.url = value

    @property
    def bytes(self) -> bytes:
        return self.proto.base64

    @bytes.setter
    def bytes(self, value: bytes):
        self.proto.base64 = value

    @classmethod
    def from_url(cls, url: str) -> "Video":
        proto_video = VideoProto(url=url)
        return cls(proto_video)

    def __repr__(self) -> str:
        attrs = []
        if self.url:
            attrs.append(f"url={self.url!r}")
        if self.bytes:
            attrs.append(f"bytes=<{len(self.bytes)} bytes>")
        return f"Video({', '.join(attrs)})"

    def to_proto(self) -> VideoProto:
        return self.proto

    def to_base64_str(self) -> str:
        if not self.proto.base64:
            raise ValueError("Video has no bytes")
        if isinstance(self.proto.base64, str):
            return self.proto.base64
        if isinstance(self.proto.base64, bytes):
            try:
                # trying direct decode (if already a base64 bytes)
                return self.proto.base64.decode('utf-8')
            except UnicodeDecodeError:
                import base64

                return base64.b64encode(self.proto.base64).decode('utf-8')
        else:
            raise TypeError("Expected str or bytes for Video.base64")

    @classmethod
    def from_proto(cls, proto: VideoProto) -> "Video":
        return cls(proto)


def cast(value, python_type):
    list_type = get_origin(python_type) is list
    if isinstance(value, MessageData):
        return value.cast(python_type)
    if list_type and isinstance(value, np.ndarray):
        return value.tolist()
    if list_type and isinstance(value, list):
        if get_args(python_type):
            inner_type = get_args(python_type)[0]
            return [cast(item, inner_type) for item in value]
        if not isinstance(value, Iterable):
            raise TypeError(f'Expected list, got {type(value)}')
    return value



"""Templates for initializing pipeline projects."""

from clarifai.versions import CLIENT_VERSION


def get_pipeline_config_template():
    """Get the config.yaml template for pipelines."""
    return """pipeline:
  id: "hello-world-pipeline"  # TODO: please fill in
  user_id: "your_user_id"  # TODO: please fill in
  app_id: "your_app_id"  # TODO: please fill in
  step_directories:
    - stepA
    - stepB
  orchestration_spec:
    argo_orchestration_spec: |
      apiVersion: argoproj.io/v1alpha1
      kind: Workflow
      metadata:
        generateName: hello-world-pipeline-
      spec:
        entrypoint: sequence
        templates:
        - name: sequence
          steps:
          - - name: step-a
              templateRef:
                name: users/your_user_id/apps/your_app_id/pipeline-steps/stepA  # TODO: please fill in
                template: users/your_user_id/apps/your_app_id/pipeline-steps/stepA  # TODO: please fill in
          - - name: step-b
              templateRef:
                name: users/your_user_id/apps/your_app_id/pipeline-steps/stepB  # TODO: please fill in
                template: users/your_user_id/apps/your_app_id/pipeline-steps/stepB  # TODO: please fill in
"""


def get_pipeline_step_config_template(step_id: str):
    """Get the config.yaml template for a pipeline step."""
    return f"""pipeline_step:
  id: "{step_id}"  # TODO: please fill in
  user_id: "your_user_id"  # TODO: please fill in
  app_id: "your_app_id"  # TODO: please fill in

pipeline_step_input_params:
  - name: input_text
    description: "Text input for processing"

build_info:
  python_version: "3.12"

pipeline_step_compute_info:
  cpu_limit: "500m"
  cpu_memory: "500Mi"
  num_accelerators: 0
"""


def get_pipeline_step_template(step_id: str):
    """Get the pipeline_step.py template for a pipeline step."""
    return f'''import argparse

import clarifai


def main():
    parser = argparse.ArgumentParser(description='{step_id} processing step.')
    parser.add_argument('--input_text', type=str, required=True, help='Text input for processing')

    args = parser.parse_args()

    print(clarifai.__version__)

    # TODO: Implement your pipeline step logic here
    print(f"{step_id} processed: {{args.input_text}}")


if __name__ == "__main__":
    main()
'''


def get_pipeline_step_requirements_template():
    """Get the requirements.txt template for pipeline steps."""
    return f'''clarifai=={CLIENT_VERSION}
# Add your pipeline step dependencies here
# Example:
# torch>=1.9.0
# transformers>=4.20.0
'''


def get_readme_template():
    """Get the README.md template for the pipeline project."""
    return """# Pipeline Project

This project contains a Clarifai pipeline with associated pipeline steps.

## Structure

```
├── config.yaml          # Pipeline configuration
├── stepA/               # First pipeline step
│   ├── config.yaml     # Step A configuration
│   ├── requirements.txt # Step A dependencies
│   └── 1/
│       └── pipeline_step.py  # Step A implementation
├── stepB/               # Second pipeline step
│   ├── config.yaml     # Step B configuration
│   ├── requirements.txt # Step B dependencies
│   └── 1/
│       └── pipeline_step.py  # Step B implementation
└── README.md           # This file
```

## Getting Started

1. **Configure the pipeline**: Edit `config.yaml` and update the TODO fields:
   - Set your `user_id` and `app_id`
   - Update the pipeline `id`
   - Modify the Argo orchestration spec as needed

2. **Configure pipeline steps**: For each step directory (stepA, stepB):
   - Edit `config.yaml` and fill in the TODO fields
   - Update `requirements.txt` with your dependencies
   - Implement your logic in `1/pipeline_step.py`

3. **Upload the pipeline**:
   ```bash
   clarifai pipeline upload config.yaml
   ```

This will:
- Upload the pipeline steps from the `step_directories`
- Create the pipeline with proper orchestration
- Link all components together

## Pipeline Steps

### stepA
TODO: Describe what stepA does

### stepB
TODO: Describe what stepB does

## Customization

- Add more pipeline steps by creating new directories and adding them to `step_directories` in `config.yaml`
- Modify the Argo orchestration spec to change the execution flow
- Update compute resources in each step's `config.yaml` as needed
"""



"""Dummy OpenAI model implementation for testing."""

import json
from typing import Any, Dict, Iterator

from clarifai.runners.models.openai_class import OpenAIModelClass


class MockOpenAIClient:
    """Mock OpenAI client for testing."""

    class Completions:
        def create(self, **kwargs):
            """Mock create method for compatibility."""
            if kwargs.get("stream", False):
                return MockCompletionStream(**kwargs)
            else:
                return MockCompletion(**kwargs)

    def __init__(self):
        self.chat = self  # Make self.chat point to self for compatibility
        self.completions = self.Completions()  # For compatibility with some clients


class MockCompletion:
    """Mock completion object that mimics the OpenAI completion response structure."""

    class Usage:
        def __init__(self, prompt_tokens, completion_tokens, total_tokens):
            self.total_tokens = total_tokens
            self.prompt_tokens = prompt_tokens
            self.completion_tokens = completion_tokens

        def to_dict(self):
            return dict(
                total_tokens=self.total_tokens,
                prompt_tokens=self.prompt_tokens,
                completion_tokens=self.completion_tokens,
            )

    class Choice:
        class Message:
            def __init__(self, content):
                self.content = content
                self.role = "assistant"

        def __init__(self, content):
            self.message = self.Message(content)
            self.finish_reason = "stop"
            self.index = 0

    def __init__(self, **kwargs):
        # Generate a simple response based on the last message
        messages = kwargs.get("messages")
        last_message = messages[-1] if messages else {"content": ""}
        response_text = f"Echo: {last_message.get('content', '')}"

        self.choices = [self.Choice(response_text)]
        self.usage = self.Usage(
            **{
                "prompt_tokens": len(str(messages)),
                "completion_tokens": len(response_text),
                "total_tokens": len(str(messages)) + len(response_text),
            }
        )

        self.id = "dummy-completion-id"
        self.created = 1234567890
        self.model = "dummy-model"

    def to_dict(self) -> Dict[str, Any]:
        """Convert the completion object to a dictionary."""
        return {
            "id": self.id,
            "created": self.created,
            "model": self.model,
            "choices": [
                {
                    "message": {"role": choice.message.role, "content": choice.message.content},
                    "finish_reason": choice.finish_reason,
                    "index": choice.index,
                }
                for choice in self.choices
            ],
            "usage": self.usage.to_dict(),
        }

    def model_dump(self):
        return self.to_dict()


class MockCompletionStream:
    """Mock completion stream that mimics the OpenAI streaming response structure."""

    class Chunk:
        class Choice:
            class Delta:
                def __init__(self, content=None):
                    self.content = content
                    self.role = "assistant" if content is None else None

            class Usage:
                def __init__(self, prompt_tokens, completion_tokens, total_tokens):
                    self.total_tokens = total_tokens
                    self.prompt_tokens = prompt_tokens
                    self.completion_tokens = completion_tokens

                def to_dict(self):
                    return dict(
                        total_tokens=self.total_tokens,
                        prompt_tokens=self.prompt_tokens,
                        completion_tokens=self.completion_tokens,
                    )

            def __init__(self, content=None, include_usage=False):
                self.delta = self.Delta(content)
                self.finish_reason = None if content else "stop"
                self.index = 0
                self.usage = (
                    self.Usage(**{"prompt_tokens": 10, "completion_tokens": 5, "total_tokens": 15})
                    if include_usage
                    else self.Usage(None, None, None)
                )

        def __init__(self, content=None, include_usage=False):
            self.choices = [self.Choice(content, include_usage)]
            self.id = "dummy-chunk-id"
            self.created = 1234567890
            self.model = "dummy-model"
            self.usage = self.choices[0].usage

        def to_dict(self) -> Dict[str, Any]:
            """Convert the chunk to a dictionary."""
            result = {
                "id": self.id,
                "created": self.created,
                "model": self.model,
                "choices": [
                    {
                        "delta": {"role": choice.delta.role, "content": choice.delta.content}
                        if choice.delta.content is not None
                        else {"role": choice.delta.role},
                        "finish_reason": choice.finish_reason,
                        "index": choice.index,
                    }
                    for choice in self.choices
                ],
            }
            if self.usage:
                result["usage"] = self.usage.to_dict()
            return result

        def model_dump(self):
            return self.to_dict()

    def __init__(self, **kwargs):
        # Generate a simple response based on the last message
        messages = kwargs.get("messages")

        last_message = messages[-1] if messages else {"content": ""}
        self.response_text = f"Echo: {last_message.get('content', '')}"
        # Create chunks that ensure the full text is included in the first chunk
        self.chunks = [
            self.response_text,  # First chunk contains the full text
            "",  # Final chunk is empty to indicate completion
        ]
        self.current_chunk = 0
        self.include_usage = kwargs.get("stream_options", {}).get("include_usage")

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_chunk < len(self.chunks):
            chunk = self.Chunk(self.chunks[self.current_chunk], self.include_usage)
            self.current_chunk += 1
            return chunk
        else:
            raise StopIteration


class DummyOpenAIModel(OpenAIModelClass):
    """Dummy OpenAI model implementation for testing."""

    client = MockOpenAIClient()
    model = "dummy-model"

    def _process_request(self, **kwargs) -> Dict[str, Any]:
        """Process a request for non-streaming responses."""
        completion_args = self._create_completion_args(kwargs)
        return self.client.chat.completions.create(**completion_args).model_dump()

    def _process_streaming_request(self, **kwargs) -> Iterator[Dict[str, Any]]:
        """Process a request for streaming responses."""
        completion_stream = self.client.chat.completions.create(**kwargs)

        for chunk in completion_stream:
            yield chunk.model_dump()

    # Override the method directly for testing
    @OpenAIModelClass.method
    def openai_stream_transport(self, req: str) -> Iterator[str]:
        """Direct implementation for testing purposes."""
        try:
            request_data = json.loads(req)
            request_data = self._create_completion_args(request_data)
            # Validate messages
            if not request_data.get("messages"):
                yield "Error: No messages provided"
                return

            for message in request_data["messages"]:
                if (
                    not isinstance(message, dict)
                    or "role" not in message
                    or "content" not in message
                ):
                    yield "Error: Invalid message format"
                    return

            for chunk in self._process_streaming_request(**request_data):
                yield json.dumps(chunk)
        except Exception as e:
            yield f"Error: {str(e)}"

    # Additional example method that could be added for specific model implementations
    @OpenAIModelClass.method
    def test_method(self, prompt: str) -> str:
        """Test method that simply echoes the input."""
        return f"Test: {prompt}"



import os

import pytest

from clarifai.client.model import Model
from clarifai.client.user import User

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]


def test_list_models():
    user = User()

    # list all public models
    public_models = user.list_models(user_id="all", show=False)
    assert len(public_models) > 0

    # list models of current user
    models = user.list_models(show=False)
    assert len(models) > 0

    # list models of an app
    models = user.list_models(user_id="clarifai", app_id="main", show=False)
    assert len(models) > 0

    # list public models of an account
    models = user.list_models(user_id="openai", show=True)
    assert len(models) > 0

    # list public models with valid params
    fmodels = user.list_models(featured_only=True, show=False)
    assert len(fmodels) < len(public_models)

    # list public models with invalid params
    with pytest.raises(Exception):
        user.list_models(user_id="openai", a=True, c=5)

    # return Clarifai Model instance
    models = user.list_models(
        user_id="openai", sort_by_name=True, show=False, return_clarifai_model=True
    )
    assert isinstance(models[0], Model)



from unittest import mock
from unittest.mock import Mock

import pytest

from clarifai.client.auth.helper import ClarifaiAuthHelper, clear_cache


@pytest.fixture(autouse=True)
def clear_caches():
    clear_cache()


def test_ui_default_url():
    default = ClarifaiAuthHelper("clarifai", "main", "fake_pat")
    assert default.ui == "https://clarifai.com"
    assert default.pat == "fake_pat"


@pytest.mark.parametrize(
    ("input_url", "expected_url"),
    (
        ("http://localhost:3002", "http://localhost:3002"),
        ("https://localhost:3002", "https://localhost:3002"),
        ("https://localhost", "https://localhost"),
        ("http://localhost", "http://localhost"),
        ("http://10.4.5.6", "http://10.4.5.6"),
        ("http://10.5.6.7", "http://10.5.6.7"),
        ("10.5.6.7:443", "https://10.5.6.7"),
        ("http://10.5.6.7:443", "http://10.5.6.7:443"),
        ("something.clarifai.com", "https://something.clarifai.com"),
        ("something.clarifai.com:443", "https://something.clarifai.com"),
        ("https://clarifai.com", "https://clarifai.com"),
    ),
)
def test_ui_urls(input_url, expected_url):
    helper = ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui=input_url)
    assert helper.ui == expected_url


def test_passing_no_schema_url_use_https_when_server_is_running():
    def raise_exception():
        return Mock()

    with mock.patch('urllib.request.urlopen', new_callable=raise_exception):
        helper = ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui="server")
        assert helper.ui == "https://server"


def test_passing_no_schema_url_detect_http_when_SSL_in_error():
    def raise_exception():
        return Mock(side_effect=Exception("Has SSL in error"))

    with mock.patch('urllib.request.urlopen', new_callable=raise_exception):
        helper = ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui="localhost:3002")
        assert helper.ui == "http://localhost:3002"


def test_passing_no_schema_url_show_error_when_not_server_running():
    def raise_exception():
        return Mock(side_effect=Exception("http_exception"))

    with mock.patch('urllib.request.urlopen', new_callable=raise_exception):
        with pytest.raises(
            Exception,
            match="Could not get a valid response from url: localhost:3002, is the API running there?",
        ):
            ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui="localhost:3002")


def test_passing_no_schema_url_require_port():
    def raise_exception():
        return Mock(side_effect=Exception("Has SSL in error"))

    with mock.patch('urllib.request.urlopen', new_callable=raise_exception):
        with pytest.raises(
            Exception, match="When providing an insecure url it must have both host:port format"
        ):
            ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui="localhost")


def test_exception_empty_user():
    ClarifaiAuthHelper("", "main", "fake_pat", validate=False)
    with pytest.raises(
        Exception,
        match="Need 'user_id' to not be empty in the query params or user CLARIFAI_USER_ID env var",
    ):
        ClarifaiAuthHelper("", "main", "fake_pat")


def test_exception_empty_pat():
    ClarifaiAuthHelper("clarifai", "main", "", validate=False)
    with pytest.raises(
        Exception,
        match="Need 'pat' or 'token' in the query params or use one of the CLARIFAI_PAT or CLARIFAI_SESSION_TOKEN env vars",
    ):
        ClarifaiAuthHelper("clarifai", "main", "")


def test_exception_path_root_cert():
    ClarifaiAuthHelper(
        "clarifai", "main", "fake_pat", root_certificates_path='fake_file.crt', validate=False
    )
    with pytest.raises(Exception, match="Root certificates path fake_file.crt does not exist"):
        ClarifaiAuthHelper("clarifai", "main", "fake_pat", root_certificates_path='fake_file.crt')



import time
from collections import OrderedDict, defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Tuple, Union

import numpy as np
from clarifai_grpc.grpc.api import resources_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict

from clarifai.client.dataset import Dataset
from clarifai.client.model import Model

try:
    import pandas as pd
except ImportError:
    raise ImportError("Can not import pandas. Please run `pip install pandas` to install it")

try:
    from loguru import logger
except ImportError:
    from ..logging import logger

MACRO_AVG = "macro_avg"


class EvalType(Enum):
    UNDEFINED = 0
    CLASSIFICATION = 1
    DETECTION = 2
    CLUSTERING = 3
    SEGMENTATION = 4
    TRACKER = 5


def get_eval_type(model_type):
    if "classifier" in model_type:
        return EvalType.CLASSIFICATION
    elif "visual-detector" in model_type:
        return EvalType.DETECTION
    elif "segmenter" in model_type:
        return EvalType.SEGMENTATION
    elif "embedder" in model_type:
        return EvalType.CLUSTERING
    elif "tracker" in model_type:
        return EvalType.TRACKER
    else:
        return EvalType.UNDEFINED


def to_file_name(x) -> str:
    return x.replace('/', '--')


@dataclass
class _BaseEvalResultHandler:
    model: Model
    eval_data: List[resources_pb2.EvalMetrics] = field(default_factory=list)

    def evaluate_and_wait(self, dataset: Dataset, eval_info: dict = None):
        from tqdm import tqdm

        dataset_id = dataset.id
        dataset_app_id = dataset.app_id
        dataset_user_id = dataset.user_id
        _ = self.model.evaluate(
            dataset_id=dataset_id,
            dataset_app_id=dataset_app_id,
            dataset_user_id=dataset_user_id,
            eval_info=eval_info,
        )
        latest_eval = self.model.list_evaluations()[0]
        excepted = 10
        desc = f"Please wait for the evaluation process between model {self.get_model_name()} and dataset {dataset_user_id}/{dataset_app_id}/{dataset_id} to complete."
        bar = tqdm(total=excepted, desc=desc, leave=False, ncols=0)
        while latest_eval.status.code in [
            status_code_pb2.MODEL_EVALUATING,
            status_code_pb2.MODEL_QUEUED_FOR_EVALUATION,
        ]:
            latest_eval = self.model.list_evaluations()[0]
            time.sleep(1)
            bar.update(1)

        if latest_eval.status.code == status_code_pb2.MODEL_EVALUATED:
            return latest_eval
        else:
            raise Exception(
                f"Model has failed to evaluate \n {latest_eval.status}.\nPlease check your dataset inputs!"
            )

    def find_eval_id(
        self, datasets: List[Dataset] = [], attempt_evaluate: bool = False, eval_info: dict = None
    ):
        list_eval_outputs = self.model.list_evaluations()
        self.eval_data = []
        for dataset in datasets:
            dataset.app_id = dataset.app_id or self.model.auth_helper.app_id
            dataset.user_id = dataset.user_id or self.model.auth_helper.user_id
            dataset_assert_msg = dataset.dataset_info
            # checking if dataset exists
            out = dataset.list_versions()
            try:
                next(iter(out))
            except Exception as e:
                if any(["CONN_DOES_NOT_EXIST" in _e for _e in e.args]):
                    raise Exception(
                        f"Dataset {dataset_assert_msg} does not exists. Please check datasets args"
                    )
                else:
                    # caused by sdk failure
                    pass
            # checking if model is evaluated with this dataset
            _is_found = False
            for each in list_eval_outputs:
                if each.status.code == status_code_pb2.MODEL_EVALUATED:
                    eval_dataset = each.ground_truth_dataset
                    # if version_id is empty -> get latest eval result of dataset,app,user id
                    if (
                        dataset.app_id == eval_dataset.app_id
                        and dataset.id == eval_dataset.id
                        and dataset.user_id == eval_dataset.user_id
                        and (
                            not dataset.version.id or dataset.version.id == eval_dataset.version.id
                        )
                    ):
                        # append to eval_data
                        self.eval_data.append(each)
                        _is_found = True
                        break

            # if not evaluated, but user wants to proceed it
            if not _is_found:
                if attempt_evaluate:
                    self.eval_data.append(self.evaluate_and_wait(dataset, eval_info=eval_info))
                # otherwise raise error
                else:
                    raise Exception(
                        f"Model {self.model.model_info.name} in app {self.model.model_info.app_id} is not evaluated yet with dataset {dataset_assert_msg}"
                    )

    @staticmethod
    def proto_to_dict(value):
        return MessageToDict(value, preserving_proto_field_name=True)

    @staticmethod
    def _f1(x: float, y: float):
        z = x + y
        return 2 * x * y / z if z else 0.0

    def _get_eval(self, index=0, **kwargs):
        logger.info(
            f"Model {self.get_model_name(pretify=True)}: retrieving {kwargs} metrics of dataset: {self.get_dataset_name_by_index(index)}"
        )
        result = self.model.get_eval_by_id(eval_id=self.eval_data[index].id, **kwargs)
        for k, v in kwargs.items():
            if v:
                getattr(self.eval_data[index], k).MergeFrom(getattr(result, k))

    def get_eval_data(self, metric_name: str, index=0):
        if metric_name == 'binary_metrics':
            if len(self.eval_data[index].binary_metrics) == 0:
                self._get_eval(index, binary_metrics=True)
        elif metric_name == 'label_counts':
            if self.proto_to_dict(self.eval_data[index].label_counts) == {}:
                self._get_eval(index, label_counts=True)
        elif metric_name == 'confusion_matrix':
            if self.eval_data[index].confusion_matrix.ByteSize() == 0:
                self._get_eval(index, confusion_matrix=True)
        elif metric_name == 'metrics_by_class':
            if len(self.eval_data[index].metrics_by_class) == 0:
                self._get_eval(index, metrics_by_class=True)
        elif metric_name == 'metrics_by_area':
            if len(self.eval_data[index].metrics_by_area) == 0:
                self._get_eval(index, metrics_by_area=True)

        return getattr(self.eval_data[index], metric_name)

    def get_threshold_index(self, threshold_list: list, selected_value: float = 0.5) -> int:
        assert 0 <= selected_value <= 1 and isinstance(selected_value, float)
        threshold_list = [round(each, 2) for each in threshold_list]

        def parse_precision(x):
            return len(str(x).split(".")[1])

        precision = parse_precision(selected_value)
        if precision > 2:
            selected_value = round(selected_value, 2)
            logger.warning("Round the selected value to .2 decimals")
        return threshold_list.index(selected_value)

    def get_dataset_name_by_index(self, index=0, pretify=True):
        out = self.eval_data[index].ground_truth_dataset
        if pretify:
            app_id = out.app_id
            dataset = out.id
            # out = f"{app_id}/{dataset}/{ver[:5]}" if ver else f"{app_id}/{dataset}"
            if self.model.model_info.app_id == app_id:
                out = dataset
            else:
                out = f"{app_id}/{dataset}"

        return out

    def get_model_name(self, pretify=True):
        model = self.model.model_info
        if pretify:
            app_id = model.app_id
            name = model.id
            ver = model.model_version.id
            model = f"{app_id}/{name}/{ver[:5]}" if ver else f"{app_id}/{name}"

        return model

    def _process_curve(
        self, data: resources_pb2.BinaryMetrics, metric_name: str, x: str, y: str
    ) -> Dict[str, Dict[str, np.array]]:
        """Postprocess curve"""
        x_arr = []
        y_arr = []
        threshold = []
        outputs = []

        def _make_df(xcol, ycol, concept_col, th_col):
            return pd.DataFrame({x: xcol, y: ycol, 'concept': concept_col, 'threshold': th_col})

        for bd in data:
            concept_id = bd.concept.id
            metric = eval(f'bd.{metric_name}')
            if metric.ByteSize() == 0:
                continue
            _x = np.array(eval(f'metric.{x}'))
            _y = np.array(eval(f'metric.{y}'))
            threshold = np.array(metric.thresholds)
            x_arr.append(_x)
            y_arr.append(_y)
            concept_cols = [concept_id for _ in range(len(_x))]
            outputs.append(_make_df(_x, _y, concept_cols, threshold))

        avg_x = np.mean(x_arr, axis=0)
        avg_y = np.mean(y_arr, axis=0)
        if np.isnan(avg_x).all():
            return None
        else:
            avg_cols = [MACRO_AVG for _ in range(len(avg_x))]
        outputs.append(_make_df(avg_x, avg_y, avg_cols, threshold))

        return pd.concat(outputs, axis=0)

    def parse_concept_ids(self, *args, **kwargs) -> List[str]:
        raise NotImplementedError

    def detailed_summary(self, *args, **kwargs):
        raise NotImplementedError

    def pr_curve(self, *args, **kwargs):
        raise NotImplementedError

    def roc_curve(self, *args, **kwargs):
        raise NotImplementedError

    def confusion_matrix(self, *args, **kwargs):
        raise NotImplementedError


@dataclass
class PlaceholderHandler(_BaseEvalResultHandler):
    def parse_concept_ids(self, *args, **kwargs) -> List[str]:
        return None

    def detailed_summary(self, *args, **kwargs):
        return None

    def pr_curve(self, *args, **kwargs):
        return None


@dataclass
class ClassificationResultHandler(_BaseEvalResultHandler):
    def parse_concept_ids(self, index=0) -> List[str]:
        eval_data = self.get_eval_data(metric_name='label_counts', index=index)
        concept_ids = [temp.concept.id for temp in eval_data.positive_label_counts]
        return concept_ids

    def detailed_summary(
        self, index=0, confidence_threshold: float = 0.5, **kwargs
    ) -> Union[None, Tuple[pd.DataFrame, pd.DataFrame]]:
        """Making detailed table per concept and for total concept

        Args:
            index (int, optional): Index of eval dataset. Defaults to 0.
            confidence_threshold (float, optional): confidence threshold. Defaults to 0.5.

        Returns:
            tuple: concepts dataframe, total dataframe
        """
        eval_data = self.get_eval_data('binary_metrics', index=index)
        summary = self.get_eval_data('summary', index=index)

        total_labeled = 0
        total_predicted = 0
        total_tp = 0
        total_fn = 0
        total_fp = 0
        metrics = []

        for bd in eval_data:
            concept_id = bd.concept.id
            if bd.precision_recall_curve.ByteSize() == 0:
                continue
            pr_th_index = self.get_threshold_index(
                list(bd.precision_recall_curve.thresholds), selected_value=confidence_threshold
            )
            roc_th_index = self.get_threshold_index(
                list(bd.roc_curve.thresholds), selected_value=confidence_threshold
            )
            if pr_th_index is None or roc_th_index is None:
                continue
            num_pos_labeled = bd.num_pos
            num_neg_labeled = bd.num_neg
            # TP/(TP+FP)
            precision = bd.precision_recall_curve.precision[pr_th_index]
            # TP/(TP+FN)
            recall = bd.precision_recall_curve.recall[pr_th_index]
            # FP/(FP+TN)
            fpr = bd.roc_curve.fpr[roc_th_index]
            # TP/(TP+FN)
            tpr = bd.roc_curve.tpr[roc_th_index]
            # TP+FN
            tp = int(tpr * num_pos_labeled)
            fn = num_pos_labeled - tp
            fp = int(fpr * num_neg_labeled)
            num_pos_pred = tp + fp
            f1 = self._f1(recall, precision)

            total_labeled += num_pos_labeled
            total_predicted += num_pos_pred
            total_fn += fn
            total_tp += tp
            total_fp += fp
            # roc auc, total labelled, predicted, tp, fn, fp, recall, precision, f1
            _d = OrderedDict(
                {
                    "Concept": concept_id,
                    "Accuracy (ROC AUC)": round(bd.roc_auc, 3),
                    "Total Labeled": num_pos_labeled,
                    "Total Predicted": num_pos_pred,
                    "True Positives": tp,
                    "False Negatives": fn,
                    "False Positives": fp,
                    "Recall": recall,
                    "Precision": precision,
                    "F1": f1,
                }
            )
            metrics.append(pd.DataFrame(_d, index=[0]))

        # If no valid data is found, return None
        if not metrics:
            return None
        # Make per concept df
        df = pd.concat(metrics, axis=0)
        # Make total df
        sum_df_total = sum(df["Total Labeled"])
        precision = sum(df.Precision * df["Total Labeled"]) / sum_df_total if sum_df_total else 0.0
        recall = sum(df.Recall * df["Total Labeled"]) / sum_df_total if sum_df_total else 0.0
        f1 = self._f1(recall, precision)
        df_total = pd.DataFrame(
            [
                [
                    'Total',
                    summary.macro_avg_roc_auc,
                    total_labeled,
                    total_predicted,
                    total_tp,
                    total_fn,
                    total_fp,
                    recall,
                    precision,
                    f1,
                ],
            ],
            columns=df.columns,
            index=[0],
        )

        return df, df_total

    def pr_curve(self, index=0, **kwargs) -> Union[None, pd.DataFrame]:
        """Making PR curve

        Args:
            index (int, optional): Index of eval dataset. Defaults to 0.

        Returns:
            dictionary: Keys are concept ids and 'macro_avg'. Values are dictionaries of {precision: np.array, recall: np.array}
        """
        eval_data = self.get_eval_data(metric_name='binary_metrics', index=index)
        outputs = self._process_curve(
            eval_data, metric_name='precision_recall_curve', x='recall', y='precision'
        )
        return outputs

    def roc_curve(self, index=0, **kwargs) -> Union[None, pd.DataFrame]:
        eval_data = self.get_eval_data(metric_name='binary_metrics', index=index)
        outputs = self._process_curve(eval_data, metric_name='roc_curve', x='tpr', y='fpr')
        return outputs

    def confusion_matrix(self, index=0, **kwargs):
        eval_data = self.get_eval_data(metric_name='confusion_matrix', index=index)
        concept_ids = self.parse_concept_ids(index)
        concept_ids.sort()
        data = np.zeros((len(concept_ids), len(concept_ids)), np.float32)
        for entry in eval_data.matrix:
            p = entry.predicted_concept.id
            a = entry.actual_concept.id
            if p in concept_ids and a in concept_ids:
                data[concept_ids.index(a), concept_ids.index(p)] = np.around(
                    entry.value, decimals=3
                )
            else:
                continue
        rownames = pd.MultiIndex.from_arrays([concept_ids], names=['Actual'])
        colnames = pd.MultiIndex.from_arrays([concept_ids], names=['Predicted'])
        df = pd.DataFrame(data, columns=colnames, index=rownames)

        return df


@dataclass
class DetectionResultHandler(_BaseEvalResultHandler):
    AREA_LIST = ["all", "medium", "small"]
    IOU_LIST = list(np.arange(0.5, 1.0, 0.1))

    def parse_concept_ids(self, index=0) -> List[str]:
        eval_data = self.get_eval_data(metric_name='metrics_by_class', index=index)
        concept_ids = [temp.concept.id for temp in eval_data]
        return concept_ids

    def detailed_summary(
        self,
        index=0,
        confidence_threshold: float = 0.5,
        iou_threshold: float = 0.5,
        area: str = "all",
        bypass_const: bool = False,
        **kwargs,
    ):
        if not bypass_const:
            assert iou_threshold in self.IOU_LIST, (
                f"Expected iou_threshold in {self.IOU_LIST}, got {iou_threshold}"
            )
            assert area in self.AREA_LIST, f"Expected area in {self.AREA_LIST}, got {area}"

        eval_data = self.get_eval_data('metrics_by_class', index=index)
        # summary = self.get_eval_data('summary', index=index)
        metrics = []
        for bd in eval_data:
            # total label
            _iou = round(bd.iou, 1)
            if not (area and bd.area_name == area) or not (
                iou_threshold and iou_threshold == _iou
            ):
                continue
            concept_id = bd.concept.id
            total = round(bd.num_tot, 3)
            # TP / (TP + FP)
            if len(bd.precision_recall_curve.precision) > 0:
                pr_th_index = self.get_threshold_index(
                    list(bd.precision_recall_curve.thresholds), selected_value=confidence_threshold
                )
                p = round(bd.precision_recall_curve.precision[pr_th_index], 3)
            else:
                p = 0
            # TP / (TP + FN)
            if len(bd.precision_recall_curve.recall) > 0:
                pr_th_index = self.get_threshold_index(
                    list(bd.precision_recall_curve.thresholds), selected_value=confidence_threshold
                )
                r = round(bd.precision_recall_curve.recall[pr_th_index], 3)
            else:
                r = 0
            tp = int(round(r * total, 0))
            fn = total - tp
            fp = float(tp) / p - tp if p else 0
            fp = int(round(fp, 1))
            f1 = self._f1(r, p)
            _d = {
                "Concept": concept_id,
                "Average Precision": round(float(bd.avg_precision), 3),
                "Total Labeled": total,
                "True Positives": tp,
                "False Positives": fp,
                "False Negatives": fn,
                "Recall": r,
                "Precision": p,
                "F1": f1,
            }
            metrics.append(pd.DataFrame(_d, index=[0]))

        if not metrics:
            return None

        df = pd.concat(metrics, axis=0)
        df_total = defaultdict()
        sum_df_total = df["Total Labeled"].sum()
        df_total["Concept"] = "Total"
        df_total["Average Precision"] = df["Average Precision"].mean()
        df_total["Total Labeled"] = sum_df_total
        df_total["True Positives"] = df["True Positives"].sum()
        df_total["False Positives"] = df["False Positives"].sum()
        df_total["False Negatives"] = df["False Negatives"].sum()
        df_total["Recall"] = (
            sum(df.Recall * df["Total Labeled"]) / sum_df_total if sum_df_total else 0.0
        )
        df_total["Precision"] = (
            df_total["True Positives"] / (df_total["True Positives"] + df_total["False Positives"])
            if sum_df_total
            else 0.0
        )
        df_total["F1"] = self._f1(df_total["Recall"], df_total["Precision"])
        df_total = pd.DataFrame(df_total, index=[0])

        return [df, df_total]

    def pr_curve(
        self, index=0, iou_threshold: float = 0.5, area: str = "all", bypass_const=False, **kwargs
    ):
        if not bypass_const:
            assert iou_threshold in self.IOU_LIST, (
                f"Expected iou_threshold in {self.IOU_LIST}, got {iou_threshold}"
            )
            assert area in self.AREA_LIST, f"Expected area in {self.AREA_LIST}, got {area}"

        eval_data = self.get_eval_data(metric_name='metrics_by_class', index=index)
        _valid_eval_data = []
        for bd in eval_data:
            _iou = round(bd.iou, 1)
            if not (area and bd.area_name == area) or not (
                iou_threshold and iou_threshold == _iou
            ):
                continue
            _valid_eval_data.append(bd)

        outputs = self._process_curve(
            _valid_eval_data, metric_name='precision_recall_curve', x='recall', y='precision'
        )
        return outputs

    def roc_curve(self, index=0, **kwargs) -> None:
        return None

    def confusion_matrix(self, index=0, **kwargs) -> None:
        return None


def make_handler_by_type(model_type: str) -> _BaseEvalResultHandler:
    _eval_type = get_eval_type(model_type)
    if _eval_type == EvalType.CLASSIFICATION:
        return ClassificationResultHandler
    elif _eval_type == EvalType.DETECTION:
        return DetectionResultHandler
    else:
        return PlaceholderHandler



import os

import pytest
from clarifai_grpc.grpc.api import resources_pb2

from clarifai.client.input import Inputs
from clarifai.client.model import Model
from clarifai.errors import UserError

DOG_IMAGE_URL = "https://samples.clarifai.com/dog2.jpeg"
NON_EXISTING_IMAGE_URL = "http://example.com/non-existing.jpg"
RED_TRUCK_IMAGE_FILE_PATH = os.path.dirname(__file__) + "/assets/red-truck.png"
BEER_VIDEO_URL = "https://samples.clarifai.com/beer.mp4"

MAIN_APP_ID = "main"
MAIN_APP_USER_ID = "clarifai"
GENERAL_MODEL_ID = "aaa03c23b3724a16a56b629203edc62c"
CLIP_EMBED_MODEL_ID = "multimodal-clip-embed"

RAW_TEXT = "Hi my name is Jim."
RAW_TEXT_BYTES = b"Hi my name is Jim."

CLARIFAI_PAT = os.environ["CLARIFAI_PAT"]
CLARIFAI_API_BASE = os.environ.get("CLARIFAI_API_BASE", "https://api.clarifai.com")


@pytest.fixture
def model():
    return Model(
        user_id=MAIN_APP_USER_ID,
        app_id=MAIN_APP_ID,
        model_id=GENERAL_MODEL_ID,
        pat=CLARIFAI_PAT,
        base_url=CLARIFAI_API_BASE,
    )


@pytest.fixture
def clip_embed_model():
    return Model(
        user_id=MAIN_APP_USER_ID,
        app_id=MAIN_APP_ID,
        model_id=CLIP_EMBED_MODEL_ID,
        pat=CLARIFAI_PAT,
        base_url=CLARIFAI_API_BASE,
    )


def validate_concepts_length(response):
    assert len(response.outputs[0].data.concepts) > 0


@pytest.mark.requires_secrets
class TestModelPredict:
    def test_predict_image_url(self, model):
        response = model.predict_by_url(DOG_IMAGE_URL)
        validate_concepts_length(response)

    def test_predict_filepath(self, model):
        response = model.predict_by_filepath(RED_TRUCK_IMAGE_FILE_PATH)
        validate_concepts_length(response)

    def test_predict_image_bytes(self, model):
        with open(RED_TRUCK_IMAGE_FILE_PATH, "rb") as f:
            image_bytes = f.read()

        response = model.predict_by_bytes(image_bytes, 'image')
        validate_concepts_length(response)

    def test_predict_image_url_with_selected_concepts(self, model):
        selected_concepts = [
            resources_pb2.Concept(name="dog"),
            resources_pb2.Concept(name="cat"),
        ]

        response = model.predict_by_url(
            DOG_IMAGE_URL, 'image', output_config=dict(select_concepts=selected_concepts)
        )
        concepts = response.outputs[0].data.concepts

        assert len(concepts) == 2
        dog_concept = next(c for c in concepts if c.name == "dog")
        cat_concept = next(c for c in concepts if c.name == "cat")
        assert dog_concept.value > cat_concept.value

    def test_predict_image_url_with_min_value(self, model):
        response = model.predict_by_url(DOG_IMAGE_URL, 'image', output_config=dict(min_value=0.98))
        assert len(response.outputs[0].data.concepts) > 0
        for c in response.outputs[0].data.concepts:
            assert c.value >= 0.98

    def test_predict_image_url_with_max_concepts(self, model):
        response = model.predict_by_url(DOG_IMAGE_URL, 'image', output_config=dict(max_concepts=3))
        assert len(response.outputs[0].data.concepts) == 3

    def test_failed_predicts(self, model):
        # Invalid FilePath
        false_filepath = "false_filepath"
        with pytest.raises(UserError):
            model.predict_by_filepath(false_filepath, 'image')

        # Invalid URL
        with pytest.raises(Exception):
            model.predict_by_url(NON_EXISTING_IMAGE_URL, 'image')

        # Invalid Input Type
        with pytest.raises(UserError):
            model.predict_by_url(DOG_IMAGE_URL, 'invalid_input_type')

    def test_predict_video_url_with_custom_sample_ms(self, model):
        video_proto = Inputs.get_input_from_url("", video_url=BEER_VIDEO_URL)
        response = model.predict([video_proto], output_config=dict(sample_ms=2000))
        # The expected time per frame is the middle between the start and the end of the frame
        # (in milliseconds).
        expected_time = 1000

        assert len(response.outputs[0].data.frames) > 0
        for frame in response.outputs[0].data.frames:
            assert frame.frame_info.time == expected_time
            expected_time += 2000

    def test_text_embed_predict_with_raw_text(self, clip_embed_model):
        clip_dim = 512
        input_text_proto = Inputs.get_input_from_bytes(
            "", text_bytes=RAW_TEXT.encode(encoding='UTF-8')
        )
        response = clip_embed_model.predict([input_text_proto])
        assert response.outputs[0].data.embeddings[0].num_dimensions == clip_dim

        response = clip_embed_model.predict([input_text_proto])
        assert response.outputs[0].data.embeddings[0].num_dimensions == clip_dim

    def test_model_load_info(self, clip_embed_model):
        assert len(clip_embed_model.kwargs) == 4
        clip_embed_model.load_info()
        assert len(clip_embed_model.kwargs) > 10



import pytest

from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.constants import DEFAULT_UI


@pytest.fixture()
def helper():
    return ClarifaiUrlHelper()


def test_module_Ui_url(helper):
    url = helper.module_ui_url("clarifai", "main", "module_1", "version_1")
    assert url == f"{DEFAULT_UI}/clarifai/main/modules/module_1/versions/version_1"


def test_module_install_url(helper):
    install_url = helper.module_install_ui_url("person", "app", "XXX")
    assert (
        install_url
        == f"{DEFAULT_UI}/person/app/installed_module_versions/module_manager_install/install?install=XXX"
    )


def test_install_with_custom_imv_id():
    custom_imv_id = "some_imv_id"
    helper = ClarifaiUrlHelper(module_manager_imv_id=custom_imv_id)
    install_url = helper.module_install_ui_url("person", "app", "XXX")
    assert (
        install_url
        == f"{DEFAULT_UI}/person/app/installed_module_versions/some_imv_id/install?install=XXX"
    )


def test_split_of_ui_clarifai_url():
    (
        user_id,
        app_id,
        module_id,
        module_version_id,
    ) = ClarifaiUrlHelper.split_module_ui_url(
        "clarifai.com/clarifai/main/modules/module_1/versions/2"
    )
    assert user_id == "clarifai"
    assert app_id == "main"
    assert module_id == "module_1"
    assert module_version_id == "2"


def test_split_with_https_prefix():
    (
        user_id,
        app_id,
        module_id,
        module_version_id,
    ) = ClarifaiUrlHelper.split_module_ui_url(
        "https://clarifai.com/clarifai/main/modules/module_1/versions/2"
    )
    assert user_id == "clarifai"
    assert app_id == "main"
    assert module_id == "module_1"
    assert module_version_id == "2"


def test_too_many_things_happen():
    with pytest.raises(
        ValueError,
        match="Provided url must have 4 or 6 parts after the domain name",
    ):
        ClarifaiUrlHelper.split_module_ui_url(
            "https://clarifai.com/clarifai/main/modules/module_1/versions/2/something"
        )


def test_not_enougth_items():
    with pytest.raises(
        ValueError, match="Provided url must have 4 or 6 parts after the domain name"
    ):
        ClarifaiUrlHelper.split_module_ui_url("clarifai/main/modules/module_1/versions/2")


def test_no_right_modules_in_path():
    with pytest.raises(
        ValueError, match="Provided url must have 4 or 6 parts after the domain name"
    ):
        ClarifaiUrlHelper.split_module_ui_url("clarifai/main/modules_abc/module_1/versions/2")


def test_no_right_module_version_in_path():
    with pytest.raises(
        ValueError, match="Provided url must have 4 or 6 parts after the domain name"
    ):
        ClarifaiUrlHelper.split_module_ui_url(
            "clarifai/main/modules/module_1/module_versions_abc/2"
        )


def test_no_right_model_version_in_path():
    with pytest.raises(
        ValueError, match="Provided url must have 4 or 6 parts after the domain name"
    ):
        ClarifaiUrlHelper.split_clarifai_url("clarifai/main/models/model_1/model_versions_abc/2")


def test_split_model_with_https_prefix():
    url = "https://clarifai.com/clarifai/main/models/model_1/versions/2"
    (
        user_id,
        app_id,
        resource_type,
        model_id,
        model_version_id,
    ) = ClarifaiUrlHelper.split_clarifai_url(url)
    assert user_id == "clarifai"
    assert app_id == "main"
    assert resource_type == "models"
    assert model_id == "model_1"
    assert model_version_id == "2"

    new = ClarifaiUrlHelper().clarifai_url(
        user_id, app_id, resource_type, model_id, model_version_id
    )
    assert new == url


def test_split_model_without_version_with_https_prefix():
    url = "https://clarifai.com/clarifai/main/models/model_1"
    (
        user_id,
        app_id,
        resource_type,
        model_id,
        model_version_id,
    ) = ClarifaiUrlHelper.split_clarifai_url(url)
    assert user_id == "clarifai"
    assert app_id == "main"
    assert resource_type == "models"
    assert model_id == "model_1"
    assert model_version_id is None

    new = ClarifaiUrlHelper().clarifai_url(
        user_id, app_id, resource_type, model_id, model_version_id
    )
    assert new == url


def test_bad_resource_type_in_clarifai_url():
    helper = ClarifaiUrlHelper()

    with pytest.raises(
        ValueError,
        match="resource_type must be one of modules, models, concepts, inputs, workflows, tasks, installed_module_versions but was models_abc",
    ):
        user_id = "clarifai"
        app_id = "main"
        resource_type = "models_abc"
        model_id = "model_1"
        model_version_id = "2"
        helper.clarifai_url(user_id, app_id, resource_type, model_id, model_version_id)



from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple, Type

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct

from clarifai.client.input import Inputs
from clarifai.utils.misc import get_uuid

from .base import ClarifaiDataLoader, ClarifaiDataset


class TextClassificationDataset(ClarifaiDataset):
    """Upload text classification datasets to clarifai datasets"""

    def __init__(
        self, data_generator: Type[ClarifaiDataLoader], dataset_id: str, max_workers: int = 4
    ) -> None:
        super().__init__(data_generator, dataset_id, max_workers)

    def _extract_protos(
        self, batch_input_ids: List[int]
    ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
        """Create input image and annotation protos for batch of input ids.
        Args:
          batch_input_ids: List of input IDs to retrieve the protos for.
        Returns:
          input_protos: List of input protos.
          annotation_protos: List of annotation protos.
        """
        input_protos, annotation_protos = [], []

        def process_data_item(id):
            data_item = self.data_generator[id]
            metadata = Struct()
            text = data_item.text
            labels = (
                data_item.labels
                if ((data_item.labels is None) or isinstance(data_item.labels, list))
                else [data_item.labels]
            )  # clarifai concept
            label_ids = data_item.label_ids
            input_id = (
                f"{self.dataset_id}-{get_uuid(8)}"
                if data_item.id is None
                else f"{self.dataset_id}-{str(data_item.id)}"
            )
            if data_item.metadata is not None:
                metadata.update(data_item.metadata)

            self.all_input_ids[id] = input_id
            input_protos.append(
                Inputs.get_text_input(
                    input_id=input_id,
                    raw_text=text,
                    dataset_id=self.dataset_id,
                    labels=labels,
                    label_ids=label_ids,
                    metadata=metadata,
                )
            )

        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures = [executor.submit(process_data_item, id) for id in batch_input_ids]
            for job in futures:
                job.result()

        return input_protos, annotation_protos



import os
from typing import Any, Dict, Generator, List

import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict

from clarifai.client.base import BaseClient
from clarifai.client.lister import Lister
from clarifai.client.nodepool import Nodepool
from clarifai.errors import UserError
from clarifai.utils.constants import DEFAULT_BASE
from clarifai.utils.logging import logger


class ComputeCluster(Lister, BaseClient):
    """ComputeCluster is a class that provides access to Clarifai API endpoints related to Compute Cluster information."""

    def __init__(
        self,
        compute_cluster_id: str = None,
        user_id: str = None,
        base_url: str = DEFAULT_BASE,
        pat: str = None,
        token: str = None,
        root_certificates_path: str = None,
        **kwargs,
    ):
        """Initializes an ComputeCluster object.

        Args:
            compute_cluster_id (str): The ComputeCluster ID for the ComputeCluster to interact with.
            user_id (str): The user ID of the user.
            base_url (str): Base API url. Default "https://api.clarifai.com"
            pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
            token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
            root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
            **kwargs: Additional keyword arguments to be passed to the compute cluster.
        """
        self.kwargs = {**kwargs, 'id': compute_cluster_id, 'user_id': user_id}
        self.compute_cluster_info = resources_pb2.ComputeCluster(**self.kwargs)
        self.logger = logger
        BaseClient.__init__(
            self,
            user_id=self.user_id,
            base=base_url,
            pat=pat,
            token=token,
            root_certificates_path=root_certificates_path,
        )
        Lister.__init__(self)

    def list_nodepools(
        self, page_no: int = None, per_page: int = None
    ) -> Generator[Nodepool, None, None]:
        """Lists all the available nodepools of the compute cluster.

        Args:
            compute_cluster_id (str): The compute cluster ID to list the nodepools.
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Nodepool: Nodepool objects for the nodepools in the compute cluster.

        Example:
            >>> from clarifai.client.compute_cluster import ComputeCluster
            >>> compute_cluster = ComputeCluster(compute_cluster_id="compute_cluster_id", user_id="user_id")
            >>> all_nodepools = list(compute_cluster.list_nodepools())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(user_app_id=self.user_app_id, compute_cluster_id=self.id)
        all_nodepools_info = self.list_pages_generator(
            self.STUB.ListNodepools,
            service_pb2.ListNodepoolsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )

        for nodepool_info in all_nodepools_info:
            yield Nodepool.from_auth_helper(auth=self.auth_helper, **nodepool_info)

    def _process_nodepool_config(self, nodepool_config: str) -> Dict[str, Any]:
        assert "nodepool" in nodepool_config, "nodepool info not found in the config file"
        nodepool = nodepool_config['nodepool']
        assert "instance_types" in nodepool, "instance_types not found in the config file"
        assert "node_capacity_type" in nodepool, "node_capacity_type not found in the config file"
        assert "min_instances" in nodepool, "min_instances not found in the config file"
        assert "max_instances" in nodepool, "max_instances not found in the config file"
        nodepool['compute_cluster'] = resources_pb2.ComputeCluster(
            id=self.id, user_id=self.user_id
        )
        nodepool['node_capacity_type'] = resources_pb2.NodeCapacityType(
            capacity_types=[
                capacity_type for capacity_type in nodepool['node_capacity_type']['capacity_types']
            ]
        )
        instance_types = []
        for instance_type in nodepool['instance_types']:
            if 'compute_info' in instance_type:
                instance_type['compute_info'] = resources_pb2.ComputeInfo(
                    **instance_type['compute_info']
                )
            instance_types.append(resources_pb2.InstanceType(**instance_type))
        nodepool['instance_types'] = instance_types
        if "visibility" in nodepool:
            nodepool["visibility"] = resources_pb2.Visibility(**nodepool["visibility"])
        return nodepool

    def create_nodepool(
        self,
        config_filepath: str = None,
        nodepool_id: str = None,
        nodepool_config: Dict[str, Any] = None,
    ) -> Nodepool:
        """Creates a nodepool for the compute cluster.

        Args:
            config_filepath (str): The path to the nodepool config file.
            nodepool_id (str): New nodepool ID for the nodepool to create.
            nodepool_config (Dict[str, Any]) = nodepool_config or {}

        Returns:
            Nodepool: A Nodepool object for the specified nodepool ID.

        Example:
            >>> from clarifai.client.compute_cluster import ComputeCluster
            >>> compute_cluster = ComputeCluster(compute_cluster_id="compute_cluster_id", user_id="user_id")
            >>> nodepool = compute_cluster.create_nodepool(config_filepath="config.yml")
        """

        if config_filepath is not None:
            assert nodepool_config is None, (
                "nodepool_config has to be None if config_filepath is provided"
            )

            if not os.path.exists(config_filepath):
                raise UserError(f"Nodepool config file not found at {config_filepath}")
            with open(config_filepath, "r") as file:
                nodepool_config = yaml.safe_load(file)
        elif nodepool_config is not None:
            assert isinstance(nodepool_config, dict), (
                "nodepool_config should be a dictionary if provided."
            )
        else:
            raise AssertionError("Either config_filepath or nodepool_config must be provided.")

        nodepool_config = self._process_nodepool_config(nodepool_config)

        if 'id' in nodepool_config:
            if nodepool_id is None:
                nodepool_id = nodepool_config['id']
            nodepool_config.pop('id')

        request = service_pb2.PostNodepoolsRequest(
            user_app_id=self.user_app_id,
            compute_cluster_id=self.id,
            nodepools=[resources_pb2.Nodepool(id=nodepool_id, **nodepool_config)],
        )
        response = self._grpc_request(self.STUB.PostNodepools, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nNodepool created\n%s", response.status)

        dict_response = MessageToDict(response.nodepools[0], preserving_proto_field_name=True)
        kwargs = self.process_response_keys(dict_response, 'nodepool')
        return Nodepool.from_auth_helper(auth=self.auth_helper, **kwargs)

    def nodepool(self, nodepool_id: str) -> Nodepool:
        """Returns a Nodepool object for the existing nodepool ID.

        Args:
            nodepool_id (str): The nodepool ID for the nodepool to interact with.

        Returns:
            Nodepool: A Nodepool object for the existing nodepool ID.

        Example:
            >>> from clarifai.client.compute_cluster import ComputeCluster
            >>> compute_cluster = ComputeCluster(compute_cluster_id="compute_cluster_id", user_id="user_id")
            >>> nodepool = compute_cluster.nodepool(nodepool_id="nodepool_id")
        """
        request = service_pb2.GetNodepoolRequest(
            user_app_id=self.user_app_id, compute_cluster_id=self.id, nodepool_id=nodepool_id
        )
        response = self._grpc_request(self.STUB.GetNodepool, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        dict_response = MessageToDict(response, preserving_proto_field_name=True)

        kwargs = self.process_response_keys(dict_response['nodepool'], 'nodepool')
        return Nodepool.from_auth_helper(auth=self.auth_helper, **kwargs)

    def delete_nodepools(self, nodepool_ids: List[str]) -> None:
        """Deletes list of nodepools for the compute cluster.

        Args:
            nodepool_ids (List[str]): The nodepool IDs of the compute cluster to delete.

        Example:
            >>> from clarifai.client.compute_cluster import ComputeCluster
            >>> compute_cluster = ComputeCluster(compute_cluster_id="compute_cluster_id", user_id="user_id")
            >>> compute_cluster.delete_nodepools(nodepool_ids=["nodepool_id1", "nodepool_id2"])
        """
        assert isinstance(nodepool_ids, list), "nodepool_ids param should be a list"

        request = service_pb2.DeleteNodepoolsRequest(
            user_app_id=self.user_app_id, compute_cluster_id=self.id, ids=nodepool_ids
        )
        response = self._grpc_request(self.STUB.DeleteNodepools, request)

        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nNodepools Deleted\n%s", response.status)

    def __getattr__(self, name):
        return getattr(self.compute_cluster_info, name)

    def __str__(self):
        init_params = [param for param in self.kwargs.keys()]
        attribute_strings = [
            f"{param}={getattr(self.compute_cluster_info, param)}"
            for param in init_params
            if hasattr(self.compute_cluster_info, param)
        ]
        return f"Clarifai Compute Cluster Details: \n{', '.join(attribute_strings)}\n"



"""Templates for model initialization."""

from clarifai import __version__


def get_model_class_template() -> str:
    """Return the template for a basic ModelClass-based model."""
    return '''from typing import Iterator, List
from clarifai.runners.models.model_class import ModelClass
from clarifai.runners.util.data_utils import Param

class MyModel(ModelClass):
    """A custom model implementation using ModelClass."""

    def load_model(self):
        """Load the model here.
        # TODO: please fill in
        # Add your model loading logic here
        """
        pass

    @ModelClass.method
    def predict(
        self,
        prompt: str = "",
        chat_history: List[dict] = None,
        max_tokens: int = Param(default=256, description="The maximum number of tokens to generate. Shorter token lengths will provide faster performance."),
        temperature: float = Param(default=1.0, description="A decimal number that determines the degree of randomness in the response"),
        top_p: float = Param(default=1.0, description="An alternative to sampling with temperature, where the model considers the results of the tokens with top_p probability mass."),
    ) -> str:
        """This is the method that will be called when the runner is run. It takes in an input and returns an output."""
        # TODO: please fill in
        # Implement your prediction logic here
        pass  # Replace with your actual logic

    @ModelClass.method
    def generate(
        self,
        prompt: str = "",
        chat_history: List[dict] = None,
        max_tokens: int = Param(default=256, description="The maximum number of tokens to generate. Shorter token lengths will provide faster performance."),
        temperature: float = Param(default=1.0, description="A decimal number that determines the degree of randomness in the response"),
        top_p: float = Param(default=1.0, description="An alternative to sampling with temperature, where the model considers the results of the tokens with top_p probability mass."),
    ) -> Iterator[str]:
        """Example yielding a streamed response."""
        # TODO: please fill in
        # Implement your generation logic here
        pass # Replace with your actual logic
'''


def get_mcp_model_class_template() -> str:
    """Return the template for an MCPModelClass-based model."""
    return '''from typing import Any

from fastmcp import FastMCP  # use fastmcp v2 not the built in mcp
from pydantic import Field

from clarifai.runners.models.mcp_class import MCPModelClass

# TODO: please fill in
# Configure your FastMCP server
server = FastMCP("my-mcp-server", instructions="", stateless_http=True)


# TODO: please fill in
# Add your tools, resources, and prompts here
@server.tool("example_tool", description="An example tool")
def example_tool(input_param: Any = Field(description="Example input parameter")):
    """Example tool implementation."""
    # TODO: please fill in
    # Implement your tool logic here
    return f"Processed: {input_param}"


# Static resource example
@server.resource("config://version")
def get_version():
    """Example static resource."""
    # TODO: please fill in
    # Return your resource data
    return "1.0.0"


@server.prompt()
def example_prompt(text: str) -> str:
    """Example prompt template."""
    # TODO: please fill in
    # Define your prompt template
    return f"Process this text: {text}"


class MyModel(MCPModelClass):
    """A custom model implementation using MCPModelClass."""

    def get_server(self) -> FastMCP:
        """Return the FastMCP server instance."""
        return server
'''


def get_openai_model_class_template() -> str:
    """Return the template for an OpenAIModelClass-based model."""
    return '''from typing import List
from openai import OpenAI
from clarifai.runners.models.openai_class import OpenAIModelClass
from clarifai.runners.util.data_utils import Param
from clarifai.runners.utils.openai_convertor import build_openai_messages

class MyModel(OpenAIModelClass):
    """A custom model implementation using OpenAIModelClass."""

    # TODO: please fill in
    # Configure your OpenAI-compatible client for local model
    client = OpenAI(
        api_key="local-key",  # TODO: please fill in - use your local API key
        base_url="http://localhost:8000/v1",  # TODO: please fill in - your local model server endpoint
    )

    # TODO: please fill in
    # Specify the model name to use
    model = "my-local-model"  # TODO: please fill in - replace with your local model name

    def load_model(self):
        """Optional: Add any additional model loading logic here."""
        # TODO: please fill in (optional)
        # Add any initialization logic if needed
        pass

    @OpenAIModelClass.method
    def predict(
        self,
        prompt: str = "",
        chat_history: List[dict] = None,
        max_tokens: int = Param(default=256, description="The maximum number of tokens to generate. Shorter token lengths will provide faster performance."),
        temperature: float = Param(default=1.0, description="A decimal number that determines the degree of randomness in the response"),
        top_p: float = Param(default=1.0, description="An alternative to sampling with temperature, where the model considers the results of the tokens with top_p probability mass."),
    ) -> str:
        """Run a single prompt completion using the OpenAI client."""
        # TODO: please fill in
        # Implement your prediction logic here
        messages = build_openai_messages(prompt, chat_history)
        response = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_completion_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
        )
        return response.choices[0].message.content

    @OpenAIModelClass.method
    def generate(
        self,
        prompt: str = "",
        chat_history: List[dict] = None,
        max_tokens: int = Param(default=256, description="The maximum number of tokens to generate. Shorter token lengths will provide faster performance."),
        temperature: float = Param(default=1.0, description="A decimal number that determines the degree of randomness in the response"),
        top_p: float = Param(default=1.0, description="An alternative to sampling with temperature, where the model considers the results of the tokens with top_p probability mass."),
    ):
        """Stream a completion response using the OpenAI client."""
        # TODO: please fill in
        # Implement your streaming logic here
        messages = build_openai_messages(prompt, chat_history)
        stream = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_completion_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            stream=True,
        )
        for chunk in stream:
            if chunk.choices:
                text = (chunk.choices[0].delta.content
                        if (chunk and chunk.choices[0].delta.content) is not None else '')
                yield text
'''


def get_config_template(model_type_id: str = "text-to-text") -> str:
    """Return the template for config.yaml."""
    return f'''# Configuration file for your Clarifai model

model:
  id: "my-model"  # TODO: please fill in - replace with your model ID
  user_id: "user_id"  # TODO: please fill in - replace with your user ID
  app_id: "app_id"  # TODO: please fill in - replace with your app ID
  model_type_id: "{model_type_id}"  # TODO: please fill in - replace if different model type ID

build_info:
  python_version: "3.12"

# TODO: please fill in - adjust compute requirements for your model
inference_compute_info:
  cpu_limit: "1"  # TODO: please fill in - Amount of CPUs to use as a limit
  cpu_memory: "1Gi"  # TODO: please fill in - Amount of CPU memory to use as a limit
  cpu_requests: "0.5"  # TODO: please fill in - Amount of CPUs to use as a minimum
  cpu_memory_requests: "512Mi"  # TODO: please fill in - Amount of CPU memory to use as a minimum
  num_accelerators: 1  # TODO: please fill in - Amount of GPU/TPUs to use
  accelerator_type: ["NVIDIA-*"]  # TODO: please fill in - type of accelerators requested
  accelerator_memory: "1Gi"  # TODO: please fill in - Amount of accelerator/GPU memory to use as a minimum

# TODO: please fill in (optional) - add checkpoints section if needed
# checkpoints:
#   type: "huggingface"  # supported type
#   repo_id: "your-model-repo"  # for huggingface
#   when: "build"  # or "runtime", "upload"
'''


def get_requirements_template(model_type_id: str = None) -> str:
    """Return the template for requirements.txt."""
    requirements = f'''# Clarifai SDK - required
clarifai>={__version__}
'''
    if model_type_id == "mcp":
        requirements += "fastmcp\n"
    elif model_type_id == "openai":
        requirements += "openai\n"
    requirements += '''
# TODO: please fill in - add your model's dependencies here
# Examples:
# torch>=2.0.0
# transformers>=4.30.0
# numpy>=1.21.0
# pillow>=9.0.0
'''
    return requirements


# Mapping of model type IDs to their corresponding templates
MODEL_TYPE_TEMPLATES = {
    "mcp": get_mcp_model_class_template,
    "openai": get_openai_model_class_template,
}


def get_model_template(model_type_id: str = None) -> str:
    """Get the appropriate model template based on model_type_id."""
    if model_type_id in MODEL_TYPE_TEMPLATES:
        return MODEL_TYPE_TEMPLATES[model_type_id]()
    return get_model_class_template()



import time
import uuid
from typing import Dict, List

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client.base import BaseClient
from clarifai.client.lister import Lister
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.constants import DEFAULT_BASE
from clarifai.utils.logging import logger


def _get_status_name(status_code: int) -> str:
    """Get the human-readable name for a status code."""
    status_mapping = {
        # Job status codes (these are the actual values based on the error message showing 64001)
        64001: "JOB_QUEUED",
        64002: "JOB_RUNNING",
        64003: "JOB_COMPLETED",
        64004: "JOB_FAILED",
        64005: "JOB_UNEXPECTED_ERROR",
        # Standard status codes
        10000: "SUCCESS",
        10010: "MIXED_STATUS",
    }
    return status_mapping.get(status_code, f"UNKNOWN_STATUS_{status_code}")


class Pipeline(Lister, BaseClient):
    """Pipeline is a class that provides access to Clarifai API endpoints related to Pipeline information."""

    def __init__(
        self,
        url: str = None,
        pipeline_id: str = None,
        pipeline_version_id: str = None,
        pipeline_version_run_id: str = None,
        user_id: str = None,
        app_id: str = None,
        nodepool_id: str = None,
        compute_cluster_id: str = None,
        log_file: str = None,
        base_url: str = DEFAULT_BASE,
        pat: str = None,
        token: str = None,
        root_certificates_path: str = None,
        **kwargs,
    ):
        """Initializes a Pipeline object.

        Args:
            url (str): The URL to initialize the pipeline object.
            pipeline_id (str): The Pipeline ID to interact with.
            pipeline_version_id (str): The Pipeline Version ID to interact with.
            pipeline_version_run_id (str): The Pipeline Version Run ID. If not provided, a UUID will be generated.
            user_id (str): The User ID that owns the pipeline.
            app_id (str): The App ID that contains the pipeline.
            nodepool_id (str): The Nodepool ID to run the pipeline on.
            compute_cluster_id (str): The Compute Cluster ID to run the pipeline on.
            log_file (str): Path to file where logs should be written. If not provided, logs are displayed on console.
            base_url (str): Base API url. Default "https://api.clarifai.com"
            pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
            token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
            root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
            **kwargs: Additional keyword arguments to be passed to the Pipeline.
        """
        if url and pipeline_id:
            raise UserError("You can only specify one of url or pipeline_id.")
        if not url and not pipeline_id:
            raise UserError("You must specify one of url or pipeline_id.")
        if url:
            parsed_user_id, parsed_app_id, _, parsed_pipeline_id, parsed_version_id = (
                ClarifaiUrlHelper.split_clarifai_url(url)
            )
            user_id = user_id or parsed_user_id
            app_id = app_id or parsed_app_id
            pipeline_id = parsed_pipeline_id
            pipeline_version_id = pipeline_version_id or parsed_version_id

        self.pipeline_id = pipeline_id
        self.pipeline_version_id = pipeline_version_id
        self.pipeline_version_run_id = pipeline_version_run_id or str(uuid.uuid4())
        self.user_id = user_id
        self.app_id = app_id
        self.nodepool_id = nodepool_id
        self.compute_cluster_id = compute_cluster_id
        self.log_file = log_file

        BaseClient.__init__(
            self,
            user_id=user_id,
            app_id=app_id,
            base=base_url,
            pat=pat,
            token=token,
            root_certificates_path=root_certificates_path,
        )
        Lister.__init__(self)

        # Set up runner selector if compute cluster and nodepool are provided
        self._runner_selector = None
        if self.compute_cluster_id and self.nodepool_id:
            from clarifai.client.nodepool import Nodepool

            self._runner_selector = Nodepool.get_runner_selector(
                user_id=self.user_id,
                compute_cluster_id=self.compute_cluster_id,
                nodepool_id=self.nodepool_id,
            )

    def run(self, inputs: List = None, timeout: int = 3600, monitor_interval: int = 10) -> Dict:
        """Run the pipeline and monitor its progress.

        Args:
            inputs (List): List of inputs to run the pipeline with. If None, runs without inputs.
            timeout (int): Maximum time to wait for completion in seconds. Default 3600 (1 hour).
            monitor_interval (int): Interval between status checks in seconds. Default 10.

        Returns:
            Dict: The pipeline run result.
        """
        # Create a new pipeline version run
        pipeline_version_run = resources_pb2.PipelineVersionRun()
        pipeline_version_run.id = self.pipeline_version_run_id

        # Set nodepools if nodepool information is available
        if self.nodepool_id and self.compute_cluster_id:
            nodepool = resources_pb2.Nodepool(
                id=self.nodepool_id,
                compute_cluster=resources_pb2.ComputeCluster(
                    id=self.compute_cluster_id, user_id=self.user_id
                ),
            )
            pipeline_version_run.nodepools.extend([nodepool])

        run_request = service_pb2.PostPipelineVersionRunsRequest()
        run_request.user_app_id.CopyFrom(self.user_app_id)
        run_request.pipeline_id = self.pipeline_id
        run_request.pipeline_version_id = self.pipeline_version_id or ""
        run_request.pipeline_version_runs.append(pipeline_version_run)

        # Add runner selector if available
        if self._runner_selector:
            run_request.runner_selector.CopyFrom(self._runner_selector)

        logger.info(f"Starting pipeline run for pipeline {self.pipeline_id}")
        response = self.STUB.PostPipelineVersionRuns(
            run_request, metadata=self.auth_helper.metadata
        )

        if response.status.code != status_code_pb2.StatusCode.SUCCESS:
            raise UserError(
                f"Failed to start pipeline run: {response.status.description}. Details: {response.status.details}"
            )

        if not response.pipeline_version_runs:
            raise UserError("No pipeline version run was created")

        pipeline_version_run = response.pipeline_version_runs[0]
        run_id = pipeline_version_run.id or self.pipeline_version_run_id

        logger.info(f"Pipeline version run created with ID: {run_id}")

        # Monitor the run
        return self._monitor_pipeline_run(run_id, timeout, monitor_interval)

    def monitor_only(self, timeout: int = 3600, monitor_interval: int = 10) -> Dict:
        """Monitor an existing pipeline run without starting a new one.

        Args:
            timeout (int): Maximum time to wait for completion in seconds. Default 3600 (1 hour).
            monitor_interval (int): Interval between status checks in seconds. Default 10.

        Returns:
            Dict: The pipeline run result.
        """
        if not self.pipeline_version_run_id:
            raise UserError("pipeline_version_run_id is required for monitoring existing runs")

        logger.info(f"Monitoring existing pipeline run with ID: {self.pipeline_version_run_id}")

        # Monitor the existing run
        return self._monitor_pipeline_run(self.pipeline_version_run_id, timeout, monitor_interval)

    def _monitor_pipeline_run(self, run_id: str, timeout: int, monitor_interval: int) -> Dict:
        """Monitor a pipeline version run until completion.

        Args:
            run_id (str): The pipeline version run ID to monitor.
            timeout (int): Maximum time to wait for completion in seconds.
            monitor_interval (int): Interval between status checks in seconds.

        Returns:
            Dict: The pipeline run result.
        """
        start_time = time.time()
        seen_logs = set()

        while time.time() - start_time < timeout:
            # Get run status
            get_run_request = service_pb2.GetPipelineVersionRunRequest()
            get_run_request.user_app_id.CopyFrom(self.user_app_id)
            get_run_request.pipeline_id = self.pipeline_id
            get_run_request.pipeline_version_id = self.pipeline_version_id or ""
            get_run_request.pipeline_version_run_id = run_id

            try:
                run_response = self.STUB.GetPipelineVersionRun(
                    get_run_request, metadata=self.auth_helper.metadata
                )

                if run_response.status.code != status_code_pb2.StatusCode.SUCCESS:
                    logger.error(f"Error getting run status: {run_response.status.description}")
                    time.sleep(monitor_interval)
                    continue

                pipeline_run = run_response.pipeline_version_run

                # Display new log entries
                self._display_new_logs(run_id, seen_logs)

                elapsed_time = time.time() - start_time
                logger.info(f"Pipeline run monitoring... (elapsed {elapsed_time:.1f}s)")

                # Check if we have orchestration status
                if (
                    hasattr(pipeline_run, 'orchestration_status')
                    and pipeline_run.orchestration_status
                ):
                    orch_status = pipeline_run.orchestration_status
                    if hasattr(orch_status, 'status') and orch_status.status:
                        status_code = orch_status.status.code
                        status_name = _get_status_name(status_code)
                        logger.info(f"Pipeline run status: {status_code} ({status_name})")

                        # Display orchestration status details if available
                        if hasattr(orch_status, 'description') and orch_status.description:
                            logger.info(f"Orchestration status: {orch_status.description}")

                        # Success codes that allow continuation: JOB_RUNNING, JOB_QUEUED
                        if status_code in [64001, 64002]:  # JOB_QUEUED, JOB_RUNNING
                            logger.info(f"Pipeline run in progress: {status_code} ({status_name})")
                            # Continue monitoring
                        # Successful terminal state: JOB_COMPLETED
                        elif status_code == 64003:  # JOB_COMPLETED
                            logger.info("Pipeline run completed successfully!")
                            return {"status": "success", "pipeline_version_run": pipeline_run}
                        # Failure terminal states: JOB_UNEXPECTED_ERROR, JOB_FAILED
                        elif status_code in [64004, 64005]:  # JOB_FAILED, JOB_UNEXPECTED_ERROR
                            logger.error(
                                f"Pipeline run failed with status: {status_code} ({status_name})"
                            )
                            return {"status": "failed", "pipeline_version_run": pipeline_run}
                        # Handle legacy SUCCESS status for backward compatibility
                        elif status_code == status_code_pb2.StatusCode.SUCCESS:
                            logger.info("Pipeline run completed successfully!")
                            return {"status": "success", "pipeline_version_run": pipeline_run}
                        elif status_code != status_code_pb2.StatusCode.MIXED_STATUS:
                            # Log other unexpected statuses but continue monitoring
                            logger.warning(
                                f"Unexpected pipeline run status: {status_code} ({status_name}). Continuing to monitor..."
                            )

            except Exception as e:
                logger.error(f"Error monitoring pipeline run: {e}")

            time.sleep(monitor_interval)

        logger.error(f"Pipeline run timed out after {timeout} seconds")
        return {"status": "timeout"}

    def _display_new_logs(self, run_id: str, seen_logs: set):
        """Display new log entries for a pipeline version run.

        Args:
            run_id (str): The pipeline version run ID.
            seen_logs (set): Set of already seen log entry IDs.
        """
        try:
            logs_request = service_pb2.ListLogEntriesRequest()
            logs_request.user_app_id.CopyFrom(self.user_app_id)
            logs_request.pipeline_id = self.pipeline_id
            logs_request.pipeline_version_id = self.pipeline_version_id or ""
            logs_request.pipeline_version_run_id = run_id
            logs_request.log_type = "pipeline.version.run"  # Set required log type
            logs_request.page = 1
            logs_request.per_page = 50

            logs_response = self.STUB.ListLogEntries(
                logs_request, metadata=self.auth_helper.metadata
            )

            if logs_response.status.code == status_code_pb2.StatusCode.SUCCESS:
                for log_entry in logs_response.log_entries:
                    # Use log entry URL or timestamp as unique identifier
                    log_id = log_entry.url or f"{log_entry.created_at.seconds}_{log_entry.message}"
                    if log_id not in seen_logs:
                        seen_logs.add(log_id)
                        log_message = f"[LOG] {log_entry.message.strip()}"

                        # Write to file if log_file is specified, otherwise log to console
                        if self.log_file:
                            with open(self.log_file, 'a', encoding='utf-8') as f:
                                f.write(log_message + '\n')
                        else:
                            logger.info(log_message)

        except Exception as e:
            logger.debug(f"Error fetching logs: {e}")



from .models.mcp_class import MCPModelClass
from .models.model_class import ModelClass
from .models.model_runner import ModelRunner
from .models.openai_class import OpenAIModelClass

__all__ = [
    "ModelRunner",
    "ModelClass",
    "MCPModelClass",
    "OpenAIModelClass",
]



import abc

import grpc
from clarifai_grpc.grpc.api import service_pb2_grpc


class V2Stub(abc.ABC):
    """Abstract base class of clarifai api rpc client stubs"""


class RpcCallable(abc.ABC):
    """Abstract base class of clarifai api rpc callables"""


# add grpc classes as subclasses of the abcs, so they also succeed in isinstance calls
def _register_classes():
    V2Stub.register(service_pb2_grpc.V2Stub)
    for name in dir(grpc):
        if name.endswith('Callable'):
            RpcCallable.register(getattr(grpc, name))
    # add grpc aio classes as subclasses of the abcs, so they also succeed in isinstance calls
    # This is needed for calling AuthorizedRPCCallable in the async stub with metadata headers
    for name in dir(grpc.aio):
        if name.endswith('Callable'):
            RpcCallable.register(getattr(grpc.aio, name))


_register_classes()



from clarifai.client.app import App
from clarifai.client.auth.register import V2Stub
from clarifai.client.auth.stub import create_stub
from clarifai.client.base import BaseClient
from clarifai.client.dataset import Dataset
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.client.model import Model
from clarifai.client.module import Module
from clarifai.client.pipeline import Pipeline
from clarifai.client.search import Search
from clarifai.client.user import User
from clarifai.client.workflow import Workflow

__all__ = [
    'V2Stub',
    'create_stub',
    'User',
    'App',
    'Model',
    'Workflow',
    'Pipeline',
    'Module',
    'Lister',
    'Dataset',
    'Inputs',
    'BaseClient',
    'Search',
]
